#
#  Copyright (C) 2007  greg Landrum
#
#   @@ All Rights Reserved @@
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#

import doctest
import unittest

from rdkit import Chem, RDConfig
from rdkit.Dbase.DbConnection import DbConnect
from rdkit.SimDivFilters import SimilarityPickers


def load_tests(loader, tests, ignore):
  """ Add the Doctests from the module """
  tests.addTests(doctest.DocTestSuite(SimilarityPickers, optionflags=doctest.ELLIPSIS))
  return tests


class TestCase(unittest.TestCase):

  def getTestData(self):
    dbName = RDConfig.RDTestDatabase
    conn = DbConnect(dbName, 'simple_mols1')
    mols = []
    for smi, ID in conn.GetData():
      mol = Chem.MolFromSmiles(str(smi))
      mol.SetProp('_Name', str(ID))
      mols.append(mol)

    # Calculate fingerprints
    probefps = []
    for mol in mols:
      fp = Chem.RDKFingerprint(mol)
      fp._id = mol.GetProp('_Name')
      probefps.append(fp)
    return probefps

  def test_GenericPicker(self):
    picker = SimilarityPickers.GenericPicker()
    self.assertRaises(NotImplementedError, picker.MakePicks)

  def test_TopNOverallPicker(self):
    probefps = self.getTestData()

    # Find topN matches
    mol = Chem.MolFromSmiles('COC')
    probeFp = Chem.RDKFingerprint(mol)

    # The picker is initialized lazy; calculation is triggered either by len
    picker = SimilarityPickers.TopNOverallPicker(numToPick=2, probeFps=[probeFp], dataSet=probefps)
    self.assertEqual(picker._picks, None)
    self.assertEqual(len(picker), 2)
    self.assertNotEqual(picker._picks, None)
    # or by addressing the elements
    picker = SimilarityPickers.TopNOverallPicker(numToPick=2, probeFps=[probeFp], dataSet=probefps)
    self.assertEqual(picker._picks, None)
    fp, score = picker[0]
    self.assertEqual(fp._id, 'ether-1')
    self.assertEqual(score, 1.0)
    self.assertNotEqual(picker._picks, None)

  def test_SpreadPicker(self):
    probefps = self.getTestData()

    mol = Chem.MolFromSmiles('COC')
    probeFp = Chem.RDKFingerprint(mol)

    # The picker is initialized lazy; calculation is triggered either by len
    picker = SimilarityPickers.SpreadPicker(numToPick=2, probeFps=[probeFp], dataSet=probefps)
    self.assertEqual(picker._picks, None)
    self.assertEqual(len(picker), 2)

    # or by addressing the elements
    picker = SimilarityPickers.SpreadPicker(numToPick=2, probeFps=[probeFp], dataSet=probefps)
    self.assertEqual(picker._picks, None)
    fp, score = picker[0]
    self.assertEqual(fp._id, 'ether-1')
    self.assertEqual(score, 1.0)


if __name__ == '__main__':  # pragma: nocover
  unittest.main()



#
#  Copyright (C) 2018 Greg Landrum
#   All Rights Reserved
#
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#
import os
import sys
import time

from rdkit import RDConfig, TestRunner

if __name__ == '__main__':
  script = 'test_list.py'
  os.chdir(RDConfig.RDCodeDir)
  t1 = time.time()
  failed, nTests = TestRunner.RunScript(script, doLongTests=False, verbose=True)
  t2 = time.time()
  TestRunner.ReportResults(script, failed, nTests, t2 - t1, verbose=True, dest=sys.stderr)
  sys.exit(len(failed))



#
#  Copyright (c) 2011, Novartis Institutes for BioMedical Research Inc.
#  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#     * Neither the name of Novartis Institutes for BioMedical Research Inc.
#       nor the names of its contributors may be used to endorse or promote
#       products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import gzip
import io
import os
import pickle
import re
import unittest

from rdkit import RDConfig, RDLogger
from rdkit.Chem import (INCHI_AVAILABLE, ForwardSDMolSupplier, MolFromMolBlock, MolFromSmiles,
                        MolToMolBlock, MolToSmiles, SanitizeMol, rdDepictor)

if INCHI_AVAILABLE:
  from rdkit.Chem import (InchiReadWriteError, InchiToInchiKey, MolBlockToInchi, MolFromInchi,
                          MolToInchi, MolToInchiKey)

COLOR_RED = '\033[31m'
COLOR_GREEN = '\033[32m'
COLOR_RESET = '\033[0m'


def inchiDiffPrefix(inchi1, inchi2):
  inchi1 = inchi1.split('/')
  inchi2 = inchi2.split('/')
  for i in range(len(inchi1) + 1):
    if i == len(inchi1):
      break
    if i == len(inchi2) or inchi1[i] != inchi2[i]:
      break
  if len(inchi1) >= i:
    return inchi1[i][0]
  else:
    return inchi2[i][0]


def inchiDiff(inchi1, inchi2):
  inchi1 = inchi1.split('/')
  inchi2 = inchi2.split('/')
  for i in range(len(inchi1) + 1):
    if i == len(inchi1):
      break
    if i == len(inchi2) or inchi1[i] != inchi2[i]:
      break
  return ('/'.join(inchi1[:i]) + COLOR_RED + '/' + '/'.join(inchi1[i:]) + '\n' + COLOR_RESET +
          '/'.join(inchi2[:i]) + COLOR_RED + '/' + '/'.join(inchi2[i:]) + COLOR_RESET)


@unittest.skipUnless(INCHI_AVAILABLE, 'Inchi support not available')
class RegressionTest(unittest.TestCase):

  def testPrechloricAcid(self):
    examples = (
      ('OCl(=O)(=O)=O', 'InChI=1S/ClHO4/c2-1(3,4)5/h(H,2,3,4,5)'),
      ('CC1=CC2=NCC(CN2C=C1)C(=O)c3ccc4cc(C)ccc4c3.OCl(=O)(=O)=O',
       'InChI=1S/C21H20N2O.ClHO4/c1-14-3-4-17-11-18(6-5-16(17)9-14)21(24)19-12-22-20-10-15(2)7-8-23(20)13-19;2-1(3,4)5/h3-11,19H,12-13H2,1-2H3;(H,2,3,4,5)'
       ),
      ('CNc1ccc2nc3ccccc3[n+](C)c2c1.[O-]Cl(=O)(=O)=O',
       'InChI=1S/C14H13N3.ClHO4/c1-15-10-7-8-12-14(9-10)17(2)13-6-4-3-5-11(13)16-12;2-1(3,4)5/h3-9H,1-2H3;(H,2,3,4,5)'
       ),
    )
    for smiles, expected in examples:
      m = MolFromSmiles(smiles)
      inchi = MolToInchi(m)
      self.assertEqual(inchi, expected)


@unittest.skipUnless(INCHI_AVAILABLE, 'Inchi support not available')
class TestCase(unittest.TestCase):

  def setUp(self):
    self.dataset = dict()
    self.dataset_inchi = dict()
    inf = gzip.open(os.path.join(RDConfig.RDCodeDir, 'Chem/test_data', 'pubchem-hard-set.sdf.gz'),
                    'r')
    self.dataset['problematic'] = ForwardSDMolSupplier(inf, sanitize=False, removeHs=False)
    with open(os.path.join(RDConfig.RDCodeDir, 'Chem/test_data', 'pubchem-hard-set.inchi'),
              'r') as intF:
      buf = intF.read().replace('\r\n', '\n').encode('latin1')
      intF.close()
    with io.BytesIO(buf) as inF:
      pkl = inF.read()
    self.dataset_inchi['problematic'] = pickle.loads(pkl, encoding='latin1')
    # disable logging
    RDLogger.DisableLog('rdApp.warning')

  def tearDown(self):
    RDLogger.EnableLog('rdApp.warning')
    RDLogger.EnableLog('rdApp.error')

  def test0InchiWritePubChem(self):
    for fp, f in self.dataset.items():
      inchi_db = self.dataset_inchi[fp]
      same, diff, reasonable = 0, 0, 0
      for m in f:
        if m is None:  # pragma: nocover
          continue
        ref_inchi = inchi_db[m.GetProp('PUBCHEM_COMPOUND_CID')]
        x, y = MolToInchi(m), ref_inchi
        if x != y:
          # print("---------------")
          # print(m.GetProp('PUBCHEM_COMPOUND_CID'))
          # print(MolToSmiles(m))
          # print(y)
          # print(x)
          if re.search(r'.[1-9]?ClO4', x) is not None:
            reasonable += 1
            continue
          SanitizeMol(m)
          if filter(lambda i: i >= 8, [len(r) for r in m.GetRingInfo().AtomRings()]):
            reasonable += 1
            continue
          # THERE ARE NO EXAMPLES FOR THE FOLLOWING (no coverage)
          # if it is because RDKit does not think the bond is stereo
          z = MolToInchi(MolFromMolBlock(MolToMolBlock(m)))
          if y != z and inchiDiffPrefix(y, z) == 'b':
            reasonable += 1
            continue
          # some warning
          try:
            MolToInchi(m, treatWarningAsError=True)
          except InchiReadWriteError as inst:
            _, error = inst.args
            if 'Metal' in error:
              reasonable += 1
              continue

          diff += 1
          print('InChI mismatch for PubChem Compound ' + m.GetProp('PUBCHEM_COMPOUND_CID'))
          print(MolToSmiles(m, True))
          print(inchiDiff(x, y))
          print()

        else:
          same += 1

      fmt = "\n{0}InChI write Summary: {1} identical, {2} suffix variance, {3} reasonable{4}"
      print(fmt.format(COLOR_GREEN, same, diff, reasonable, COLOR_RESET))
      self.assertEqual(same, 1162)
      self.assertEqual(diff, 0)
      self.assertEqual(reasonable, 19)

  def test1InchiReadPubChem(self):
    for f in self.dataset.values():
      same, diff, reasonable = 0, 0, 0
      for m in f:
        if m is None:  # pragma: nocover
          continue
        x = MolToInchi(m)
        y = None
        RDLogger.DisableLog('rdApp.error')
        mol = MolFromInchi(x)
        RDLogger.EnableLog('rdApp.error')
        if mol is not None:
          y = MolToInchi(MolFromSmiles(MolToSmiles(mol, isomericSmiles=True)))
        if y is None:
          # metal involved?
          try:
            MolToInchi(m, treatWarningAsError=True)
          except InchiReadWriteError as inst:
            _, error = inst.args
            if 'Metal' in error or \
                    'Charges were rearranged' in error:
              reasonable += 1
              continue
          # THERE ARE NO EXAMPLES FOR THE FOLLOWING (no coverage)
          # RDKit does not like the SMILES? use MolBlock instead
          inchiMol = MolFromInchi(x)
          if inchiMol:
            rdDepictor.Compute2DCoords(inchiMol)
            z = MolToInchi(MolFromMolBlock(MolToMolBlock(inchiMol)))
            if x == z:
              reasonable += 1
              continue
          # InChI messed up the radical?
          unsanitizedInchiMol = MolFromInchi(x, sanitize=False)
          if sum([
              a.GetNumRadicalElectrons() * a.GetAtomicNum()
              for a in m.GetAtoms() if a.GetNumRadicalElectrons() != 0
          ]) != sum([
              a.GetNumRadicalElectrons() * a.GetAtomicNum()
              for a in unsanitizedInchiMol.GetAtoms() if a.GetNumRadicalElectrons() != 0
          ]):
            reasonable += 1
            continue

          diff += 1
          cid = m.GetProp('PUBCHEM_COMPOUND_CID')
          print(COLOR_GREEN + 'Empty mol for PubChem Compound ' + cid + '\n' + COLOR_RESET)
          continue
        if x != y:
          # if there was warning in the first place, then this is
          # tolerable
          try:
            MolToInchi(m, treatWarningAsError=True)
            MolFromInchi(x, treatWarningAsError=True)
          except InchiReadWriteError as inst:
            reasonable += 1
            continue
          # or if there are big rings
          SanitizeMol(m)
          if filter(lambda i: i >= 8, [len(r) for r in m.GetRingInfo().AtomRings()]):
            reasonable += 1
            continue
          # THERE ARE NO EXAMPLES FOR THE FOLLOWING (no coverage)
          # or if RDKit loses bond stereo
          s = MolToSmiles(m, True)
          if MolToSmiles(MolFromSmiles(s), True) != s:
            reasonable += 1
            continue
          # or if it is RDKit SMILES writer unhappy about the mol
          inchiMol = MolFromInchi(x)
          rdDepictor.Compute2DCoords(inchiMol)
          z = MolToInchi(MolFromMolBlock(MolToMolBlock(inchiMol)))
          if x == z:
            reasonable += 1
            continue

          diff += 1
          print(COLOR_GREEN + 'Molecule mismatch for PubChem Compound ' + cid + COLOR_RESET)
          print(inchiDiff(x, y))
          print()
        else:
          same += 1
      fmt = "\n{0}InChI read Summary: {1} identical, {2} variance, {3} reasonable variance{4}"
      print(fmt.format(COLOR_GREEN, same, diff, reasonable, COLOR_RESET))
      self.assertEqual(same, 685)
      self.assertEqual(diff, 0)
      self.assertEqual(reasonable, 496)

  def test2InchiOptions(self):
    m = MolFromSmiles("CC=C(N)C")
    inchi1 = MolToInchi(m).split('/', 1)[1]
    inchi2 = MolToInchi(m, "/SUU").split('/', 1)[1]
    self.assertEqual(inchi1 + '/b4-3?', inchi2)

  def test3InchiKey(self):
    inchi = 'InChI=1S/C9H12/c1-2-6-9-7-4-3-5-8-9/h3-5,7-8H,2,6H2,1H3'
    self.assertEqual(InchiToInchiKey(inchi), 'ODLMAHJVESYWTB-UHFFFAOYSA-N')

  def test4MolToInchiKey(self):
    m = MolFromSmiles("CC=C(N)C")
    inchi = MolToInchi(m)
    k1 = InchiToInchiKey(inchi)
    k2 = MolToInchiKey(m)
    self.assertEqual(k1, k2)

  def test5MolBlockToInchi(self):
    mb = """
  Mrv1824 02111920092D          

  6  6  0  0  0  0            999 V2000
   -5.5134    3.5259    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -6.2279    3.1134    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -6.2279    2.2884    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -5.5134    1.8759    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -4.7989    2.2884    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -4.7989    3.1134    0.0000 O   0  0  0  0  0  0  0  0  0  0  0  0
  1  2  1  0  0  0  0
  3  4  1  0  0  0  0
  4  5  1  0  0  0  0
  5  6  1  0  0  0  0
  1  6  1  0  0  0  0
  2  3  2  0  0  0  0
M  END"""
    inchi = MolBlockToInchi(mb)
    self.assertEqual(inchi, "InChI=1S/C5H8O/c1-2-4-6-5-3-1/h1-2H,3-5H2")
    # make sure that options work
    mb2 = """
  Mrv1824 02121905282D          

 10 11  0  0  0  0            999 V2000
   -4.6875   -1.1393    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -5.4020   -1.5518    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -5.4020   -2.3768    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -4.6875   -2.7893    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -3.9730   -2.3768    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -3.9730   -1.5518    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -3.2586   -2.7893    0.0000 N   0  0  0  0  0  0  0  0  0  0  0  0
   -2.5441   -1.5518    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -2.5441   -2.3768    0.0000 N   0  0  0  0  0  0  0  0  0  0  0  0
   -1.9608   -0.9684    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
  1  2  2  0  0  0  0
  2  3  1  0  0  0  0
  3  4  2  0  0  0  0
  4  5  1  0  0  0  0
  5  6  2  0  0  0  0
  1  6  1  0  0  0  0
  7  9  1  0  0  0  0
  6  8  1  0  0  0  0
  7  5  1  0  0  0  0
  8 10  1  0  0  0  0
  8  9  2  0  0  0  0
M  END"""
    inchi2 = MolBlockToInchi(mb2, options="/FixedH")
    self.assertEqual(inchi2, "InChI=1/C8H8N2/c1-6-7-4-2-3-5-8(7)10-9-6/h2-5H,1H3,(H,9,10)/f/h10H")


if __name__ == '__main__':  # pragma: nocover
  # only run the test if InChI is available
  if INCHI_AVAILABLE:
    unittest.main()



#
#  Copyright (C) 2024 Greg Landrum and other RDKit contributors
#   @@ All Rights Reserved @@
#
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.

import unittest

#
from rdkit import Chem
from rdkit.Chem import rdMolProcessing
from rdkit.Chem import rdFingerprintGenerator
from rdkit import DataStructs
from rdkit import RDConfig


class TestCase(unittest.TestCase):

  def setUp(self):
    self.smiFile = RDConfig.RDBaseDir + '/Regress/Data/zinc.leads.500.q.smi'
    self.sdFile = RDConfig.RDBaseDir + "/Data/NCI/first_200.props.sdf"

  def test1(self):
    fpg = rdFingerprintGenerator.GetMorganGenerator()
    fps = rdMolProcessing.GetFingerprintsForMolsInFile(self.smiFile)
    self.assertEqual(len(fps), 499)
    with Chem.SmilesMolSupplier(self.smiFile, delimiter='\t') as suppl:
      mols = [next(suppl) for _ in range(3)]
    nfps = [fpg.GetFingerprint(m) for m in mols]
    self.assertEqual(DataStructs.TanimotoSimilarity(fps[0], fps[1]),
                     DataStructs.TanimotoSimilarity(nfps[0], nfps[1]))

    fps = rdMolProcessing.GetFingerprintsForMolsInFile(self.sdFile)
    self.assertEqual(len(fps), 200)
    with Chem.SDMolSupplier(self.sdFile) as suppl:
      mols = [next(suppl) for _ in range(3)]
    nfps = [fpg.GetFingerprint(m) for m in mols]
    self.assertAlmostEqual(DataStructs.TanimotoSimilarity(fps[0], fps[1]), 0.0638, places=3)

  def test2(self):
    fpg = rdFingerprintGenerator.GetMorganGenerator(radius=2)

    fps = rdMolProcessing.GetFingerprintsForMolsInFile(self.smiFile, generator=fpg)
    self.assertEqual(len(fps), 499)
    with Chem.SmilesMolSupplier(self.smiFile, delimiter='\t') as suppl:
      mols = [next(suppl) for _ in range(3)]
    nfps = [fpg.GetFingerprint(m) for m in mols]
    self.assertEqual(DataStructs.TanimotoSimilarity(fps[0], fps[1]),
                     DataStructs.TanimotoSimilarity(nfps[0], nfps[1]))


if __name__ == '__main__':  # pragma: nocover
  unittest.main()



#
#  Copyright (C) 2001  greg Landrum
#
""" unit testing code for MLData sets

"""
import contextlib
import pickle
import random
import unittest
from io import StringIO

from rdkit import RDConfig
from rdkit.ML.Data import DataUtils, MLData


class TestCase(unittest.TestCase):

  def setUpQuantLoad(self):
    self.d = DataUtils.BuildQuantDataSet(RDConfig.RDCodeDir + '/ML/Data/test_data/test.qdat')

  def setUpGeneralLoad(self):
    self.d = DataUtils.BuildDataSet(RDConfig.RDCodeDir + '/ML/Data/test_data/test.dat')

  def testQuantLoad(self):
    # " testing QuantDataSet load"
    self.d = DataUtils.BuildQuantDataSet(RDConfig.RDCodeDir + '/ML/Data/test_data/test.qdat')

  def testQuantProps(self):
    # " testing QuantDataSet Properties"
    self.setUpQuantLoad()
    assert self.d.GetNPts() == 5, 'nPts wrong'
    assert self.d.GetNVars() == 4, 'nVars wrong'
    assert self.d.GetNResults() == 1, 'nResults wrong'
    assert self.d.GetVarNames() == ['foo1', 'foo2', 'foo3', 'foo4', 'res'], 'varNames wrong'
    assert self.d.GetPtNames() == ['p1', 'p2', 'p3', 'p4', 'p5'], 'ptNames wrong'
    assert self.d.GetNPossibleVals() == [3, 3, 2, 2, 2], 'nPossible Wrong'
    assert self.d.GetQuantBounds() == [[], [], [], [], []], 'quantBounds Wrong'
    assert self.d.GetResults() == [0, 0, 1, 1, 1], 'GetResults wrong'
    assert self.d.GetAllData()[1] == [0, 0, 0, 1, 0], 'GetAllData wrong'
    assert self.d.GetInputData()[3] == [2, 1, 0, 0], 'GetInputData wrong'
    assert self.d.GetNamedData()[2] == ['p3', 1, 0, 0, 0, 1], 'GetNamedData wrong'

  def testQuantPickle(self):
    # " testing QuantDataSet pickling "
    self.setUpQuantLoad()
    DataUtils.WritePickledData(RDConfig.RDCodeDir + '/ML/Data/test_data/testquant.qdat.pkl', self.d)
    with open(RDConfig.RDCodeDir + '/ML/Data/test_data/testquant.qdat.pkl', 'rb') as f:
      vNames = pickle.load(f)
      qBounds = pickle.load(f)
      ptNames = pickle.load(f)
      examples = pickle.load(f)
    d = MLData.MLQuantDataSet(examples, varNames=vNames, qBounds=qBounds, ptNames=ptNames)
    assert self.d.GetNPts() == d.GetNPts(), 'nPts wrong'
    assert self.d.GetNVars() == d.GetNVars(), 'nVars wrong'
    assert self.d.GetNResults() == d.GetNResults(), 'nResults wrong'
    assert self.d.GetVarNames() == d.GetVarNames(), 'varNames wrong'
    assert self.d.GetPtNames() == d.GetPtNames(), 'ptNames wrong'
    assert self.d.GetNPossibleVals() == d.GetNPossibleVals(), 'nPossible Wrong'
    assert self.d.GetQuantBounds() == d.GetQuantBounds(), 'quantBounds Wrong'
    assert self.d.GetResults() == d.GetResults(), 'GetResults wrong'
    assert self.d.GetAllData()[1] == d.GetAllData()[1], 'GetAllData wrong'
    assert self.d.GetInputData()[3] == d.GetInputData()[3], 'GetInputData wrong'
    assert self.d.GetNamedData()[2] == d.GetNamedData()[2], 'GetNamedData wrong'

  def testGeneralLoad(self):
    # " testing DataSet load"
    self.d = DataUtils.BuildDataSet(RDConfig.RDCodeDir + '/ML/Data/test_data/test.dat')

  def testGeneralProps(self):
    # " testing DataSet properties"
    self.setUpGeneralLoad()
    assert self.d.GetNPts() == 5, 'nPts wrong'
    assert self.d.GetNVars() == 4, 'nVars wrong'
    assert self.d.GetNResults() == 1, 'nResults wrong'
    assert self.d.GetVarNames() == ['foo1', 'foo2', 'foo3', 'foo4', 'res'], 'varNames wrong'
    assert self.d.GetPtNames() == ['p1', 'p2', 'p3', 'p4', 'p5'], 'ptNames wrong'
    assert self.d.GetNPossibleVals() == [0, 6, 0, 0, 3], 'nPossible Wrong'
    assert self.d.GetQuantBounds() == [[], [], [], [], [2, 4]], 'quantBounds Wrong'
    assert self.d.GetResults() == [1.1, 2.1, 3.1, 4.1, 5.1], 'GetResults wrong'
    assert self.d.GetAllData()[1] == ['foo', 2, 1.0, 1, 2.1], 'GetAllData wrong'
    assert self.d.GetInputData()[4] == ['foo', 5, 1.1, 1], 'GetInputData wrong'
    assert self.d.GetNamedData()[3] == ['p4', 'foo', 4, 1.0, 1, 4.1], 'GetNamedData wrong'
    assert self.d[1] == ['p2', 'foo', 2, 1.0, 1, 2.1]
    assert self.d[3] == ['p4', 'foo', 4, 1.0, 1, 4.1]
    self.d[3] = self.d[1]
    assert self.d[3] == ['p2', 'foo', 2, 1.0, 1, 2.1]

  def testGeneralPickle(self):
    # " testing DataSet pickling"
    self.setUpGeneralLoad()
    DataUtils.WritePickledData(RDConfig.RDCodeDir + '/ML/Data/test_data/testgeneral.dat.pkl',
                               self.d)
    with open(RDConfig.RDCodeDir + '/ML/Data/test_data/testgeneral.dat.pkl', 'rb') as f:
      vNames = pickle.load(f)
      qBounds = pickle.load(f)
      ptNames = pickle.load(f)
      examples = pickle.load(f)
    d = MLData.MLDataSet(examples, varNames=vNames, qBounds=qBounds, ptNames=ptNames)
    assert self.d.GetNPts() == d.GetNPts(), 'nPts wrong'
    assert self.d.GetNVars() == d.GetNVars(), 'nVars wrong'
    assert self.d.GetNResults() == d.GetNResults(), 'nResults wrong'
    assert self.d.GetVarNames() == d.GetVarNames(), 'varNames wrong'
    assert self.d.GetPtNames() == d.GetPtNames(), 'ptNames wrong'
    assert self.d.GetNPossibleVals() == d.GetNPossibleVals(), 'nPossible Wrong'
    assert self.d.GetQuantBounds() == d.GetQuantBounds(), 'quantBounds Wrong'
    assert self.d.GetResults() == d.GetResults(), 'GetResults wrong'
    assert self.d.GetAllData()[1] == d.GetAllData()[1], 'GetAllData wrong'
    assert self.d.GetInputData()[3] == d.GetInputData()[3], 'GetInputData wrong'
    assert self.d.GetNamedData()[2] == d.GetNamedData()[2], 'GetNamedData wrong'

  def test_WriteData(self):
    self.setUpQuantLoad()
    with contextlib.closing(StringIO()) as f:
      DataUtils.WriteData(f, self.d.GetVarNames(), self.d.GetQuantBounds(), self.d.data)
      s = f.getvalue()
      self.assertIn('DataUtils', s)
      self.assertIn('foo1', s)
      self.assertIn('2 2 1 0 1', s)

  def test_CalcNPossibleUsingMap(self):
    self.setUpQuantLoad()
    order = list(range(5))
    self.assertEqual(DataUtils.CalcNPossibleUsingMap(self.d.data, order, self.d.GetQuantBounds()),
                     [3, 3, 2, 2, 2])

  def test_RandomizeActivities(self):

    class RunDetails(object):
      shuffled = False
      randomized = False

    random.seed(0)
    details = RunDetails()
    self.setUpGeneralLoad()
    dataSet = self.d
    orgActivities = [d[-1] for d in dataSet]
    DataUtils.RandomizeActivities(dataSet, shuffle=True, runDetails=details)
    self.assertNotEqual(orgActivities, [d[-1] for d in dataSet])
    self.assertEqual(sorted(orgActivities), sorted([d[-1] for d in dataSet]))
    self.assertTrue(details.shuffled)
    self.assertFalse(details.randomized)

    try:
      details = RunDetails()
      self.setUpGeneralLoad()
      dataSet = self.d
      orgActivities = [d[-1] for d in dataSet]
      DataUtils.RandomizeActivities(dataSet, shuffle=False, runDetails=details)
      self.assertNotEqual(orgActivities, [d[-1] for d in dataSet])
      self.assertEqual(sorted(orgActivities), sorted([d[-1] for d in dataSet]))
      self.assertFalse(details.randomized)
      self.assertTrue(details.shuffled)
    except NameError:
      # This code branch is not working.
      pass


if __name__ == '__main__':  # pragma: nocover
  unittest.main()



# $Id$
#
#  Copyright (C) 2002-2006  greg Landrum and Rational Discovery LLC
#
#   @@ All Rights Reserved @@
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#
"""unit testing code for the database utilities

"""
import os
import tempfile
import unittest
from io import StringIO

from rdkit import RDConfig
from rdkit.Dbase import DbUtils
from rdkit.Dbase.DbConnection import DbConnect

try:
  from contextlib import redirect_stdout
except Exception:
  from rdkit.TestRunner import redirect_stdout


class TestCase(unittest.TestCase):

  def setUp(self):
    self.baseDir = os.path.join(RDConfig.RDCodeDir, 'Dbase', 'test_data')
    self.dbName = RDConfig.RDTestDatabase
    if RDConfig.useSqlLite:
      fd, tempName = tempfile.mkstemp(suffix='sqlt')
      self.fd = fd
      self.tempDbName = tempName
    else:
      self.tempDbName = '::RDTests'
    self.colHeads = ('int_col', 'floatCol', 'strCol')
    self.colTypes = ('integer', 'float', 'string')

  def tearDown(self):
    if RDConfig.useSqlLite and os.path.exists(self.tempDbName):
      os.close(self.fd)
      try:
        os.unlink(self.tempDbName)
      except Exception:
        import traceback
        traceback.print_exc()

  def _confirm(self, tblName, dbName=None, colHeads=None, colTypes=None):
    dbName = dbName or self.dbName
    colHeads = colHeads or self.colHeads
    colTypes = colTypes or self.colTypes

    conn = DbConnect(dbName, tblName)
    res = conn.GetColumnNamesAndTypes()
    assert len(res) == len(colHeads), 'bad number of columns'
    names = [x[0] for x in res]
    for i in range(len(names)):
      assert names[i].upper() == colHeads[i].upper(), 'bad column head'
    if RDConfig.useSqlLite:
      # doesn't seem to be any column type info available
      return
    types = [x[1] for x in res]
    for i in range(len(types)):
      assert types[i] == colTypes[i], 'bad column type'

  def test1Txt(self):
    """ test reading from a text file """
    with open(os.path.join(self.baseDir, 'dbtest.csv'), 'r') as inF:
      tblName = 'fromtext'
      f = StringIO()
      with redirect_stdout(f):
        DbUtils.TextFileToDatabase(self.tempDbName, tblName, inF)
      self._confirm(tblName, dbName=self.tempDbName)

  def test3Txt(self):
    """ test reading from a text file including null markers"""
    with open(os.path.join(self.baseDir, 'dbtest.nulls.csv'), 'r') as inF:
      tblName = 'fromtext2'
      f = StringIO()
      with redirect_stdout(f):
        DbUtils.TextFileToDatabase(self.tempDbName, tblName, inF, nullMarker='NA')

      self._confirm(tblName, dbName=self.tempDbName)

  def testGetData1(self):
    """ basic functionality
    """
    d = DbUtils.GetData(self.dbName, 'ten_elements', forceList=1)
    assert len(d) == 10
    assert tuple(d[0]) == (0, 11)
    assert tuple(d[2]) == (4, 31)
    with self.assertRaisesRegex(IndexError, ""):
      d[11]

  def testGetData2(self):
    """ using a RandomAccessDbResultSet
    """
    d = DbUtils.GetData(self.dbName, 'ten_elements', forceList=0, randomAccess=1)
    assert tuple(d[0]) == (0, 11)
    assert tuple(d[2]) == (4, 31)
    assert len(d) == 10
    with self.assertRaisesRegex(IndexError, ""):
      d[11]

  def testGetData3(self):
    """ using a DbResultSet
    """
    d = DbUtils.GetData(self.dbName, 'ten_elements', forceList=0, randomAccess=0)
    with self.assertRaisesRegex(TypeError, ""):
      len(d)
    rs = []
    for thing in d:
      rs.append(thing)
    assert len(rs) == 10
    assert tuple(rs[0]) == (0, 11)
    assert tuple(rs[2]) == (4, 31)

  def testGetData4(self):
    """ using a RandomAccessDbResultSet with a Transform
    """

    def fn(x):
      return (x[0], x[1] * 2)

    d = DbUtils.GetData(self.dbName, 'ten_elements', forceList=0, randomAccess=1, transform=fn)
    assert tuple(d[0]) == (0, 22)
    assert tuple(d[2]) == (4, 62)
    assert len(d) == 10
    with self.assertRaisesRegex(IndexError, ""):
      d[11]

  def testGetData5(self):
    """ using a DbResultSet with a Transform
    """

    def fn(x):
      return (x[0], x[1] * 2)

    d = DbUtils.GetData(self.dbName, 'ten_elements', forceList=0, randomAccess=0, transform=fn)
    with self.assertRaisesRegex(TypeError, ""):
      len(d)

    rs = []
    for thing in d:
      rs.append(thing)
    assert len(rs) == 10
    assert tuple(rs[0]) == (0, 22)
    assert tuple(rs[2]) == (4, 62)

  def test_take(self):
    self.assertEqual(list(DbUtils._take([1, 2, 3, 4], [2, 3])), [3, 4])
    self.assertEqual(list(DbUtils._take([1, 2, 3, 4], [0, 3])), [1, 4])

  def test_GetColumns(self):
    d = DbUtils.GetColumns(self.dbName, 'ten_elements', 'val')
    self.assertEqual(len(d), 10)

  def test_GetData_where(self):
    d = DbUtils.GetData(self.dbName, 'ten_elements_dups', forceList=0, randomAccess=0,
                        whereString='id<4')
    self.assertEqual(len(list(d)), 4)
    self.assertTrue(all(x[0] < 4 for x in d))

    d = DbUtils.GetData(self.dbName, 'ten_elements_dups', forceList=0, randomAccess=0,
                        whereString='id<10')
    self.assertEqual(len(list(d)), 10)
    self.assertTrue(all(x[0] < 10 for x in d))

    d = DbUtils.GetData(self.dbName, 'ten_elements_dups', removeDups=1, forceList=True)
    self.assertEqual(len(list(d)), 10)

  def test_DatabaseToText(self):
    txt = DbUtils.DatabaseToText(self.dbName, 'ten_elements')
    self.assertIn('id,val', txt)
    self.assertIn('0,11', txt)
    self.assertIn('18,101', txt)
    self.assertEqual(len(txt.split('\n')), 11)

    txt = DbUtils.DatabaseToText(self.dbName, 'ten_elements', fields='val')
    self.assertNotIn('id', txt)
    self.assertNotIn(',', txt)

    txt = DbUtils.DatabaseToText(self.dbName, 'ten_elements', where='id<4')
    self.assertIn('id,val', txt)
    self.assertEqual(len(txt.split('\n')), 3)

  def test_TypeFinder(self):
    data = [('-', 1.45, 'abc', None), (20, 3, 'defgh', None)]
    self.assertEqual(DbUtils.TypeFinder(data, 2, 4, nullMarker='-'),
                     [[int, 2], [float, 4], [str, 5], [-1, 1]])

  def test_AdjustColHeadings(self):
    headers = ['abc def', ' abc def', 'abc-def ', 'abc.def']
    self.assertEqual(DbUtils._AdjustColHeadings(headers, 7), ['abc_def'] * 4)
    f = StringIO()
    with redirect_stdout(f):
      headers = ['abc def', ' abc def', 'abc-def ', 'abc.def']
      self.assertEqual(DbUtils._AdjustColHeadings(headers, 3), ['abc'] * 4)
    self.assertIn('Heading', f.getvalue())

  def test_GetTypeStrings(self):
    headers = ['pk', 'a', 'b', 'c']
    colTypes = [(int, 2), (int, 3), (float, 5), (str, 10)]
    self.assertEqual(DbUtils.GetTypeStrings(headers, colTypes),
                     ['pk integer', 'a integer', 'b double precision', 'c varchar(10)'])
    self.assertEqual(
      DbUtils.GetTypeStrings(headers, colTypes, keyCol='pk'),
      ['pk integer not null primary key', 'a integer', 'b double precision', 'c varchar(10)'])

  def test_DatabaseToDatabase(self):
    tblName = 'db2db'
    f = StringIO()
    with redirect_stdout(f):
      DbUtils.DatabaseToDatabase(self.dbName, 'ten_elements', self.tempDbName, tblName)
    self._confirm(tblName, dbName=self.tempDbName, colHeads=['id', 'val'])


if __name__ == '__main__':  # pragma: nocover
  unittest.main()



"""
Tests for rdkit.Chem.RegistrationHash

Focus is on molecules that have differing SMILES, but
are actually the same.
"""

import os
import unittest

from rdkit import Chem
from rdkit import RDConfig
from rdkit.Chem import RegistrationHash
from rdkit.Chem.RegistrationHash import HashLayer


def hash_sdf(molblock: str, escape=None, data_field_names=None, enable_tautomer_hash_v2=False):
  """
    Gets the layers of the SDF, and generates the hash based on only the layers passed in

    :param molblock: The molblock to hash
    :param escape: optional field which can contain arbitrary information
    :param data_field_names: optional sequence of names of SGroup DAT fields which
       will be included in the hash.

    :return: A dict with the hash & the layers
    """
  mol = Chem.MolFromMolBlock(molblock)
  return RegistrationHash.GetMolLayers(mol, escape=escape, data_field_names=data_field_names,
                                       enable_tautomer_hash_v2=enable_tautomer_hash_v2)


class CanonicalizerTest(unittest.TestCase):
  maxDiff = 2000  # hash diffs can be long!

  def test_example_structure(self):

    structure = """
     RDKit          2D

  0  0  0  0  0  0  0  0  0  0999 V3000
M  V30 BEGIN CTAB
M  V30 COUNTS 21 23 0 0 0
M  V30 BEGIN ATOM
M  V30 1 C -2.00838 1.55562 0 0
M  V30 2 C -2.00918 0.555619 0 0
M  V30 3 C -2.87558 0.056219 0 0
M  V30 4 C -2.87638 -0.943781 0 0
M  V30 5 C -3.74278 -1.44298 0 0
M  V30 6 N -2.01078 -1.44438 0 0
M  V30 7 C -1.14438 -0.945181 0 0
M  V30 8 C -0.278781 -1.44578 0 0
M  V30 9 C 0.587619 -0.946381 0 0
M  V30 10 N 1.45322 -1.44718 0 0
M  V30 11 C 2.31962 -0.947781 0 0
M  V30 12 O 3.18522 -1.44858 0 0
M  V30 13 C 2.32042 0.0522191 0 0
M  V30 14 C 1.45482 0.552819 0 0
M  V30 15 C 1.45562 1.55282 0 0
M  V30 16 F 2.45562 1.55202 0 0
M  V30 17 F 1.95642 2.41842 0 0
M  V30 18 F 0.590019 2.05362 0 0
M  V30 19 C 0.588419 0.0536191 0 0
M  V30 20 C -0.277181 0.554219 0 0
M  V30 21 C -1.14358 0.054819 0 0
M  V30 END ATOM
M  V30 BEGIN BOND
M  V30 1 1 2 1 CFG=3
M  V30 2 1 2 3
M  V30 3 1 3 4
M  V30 4 1 4 5 CFG=3
M  V30 5 1 4 6
M  V30 6 1 6 7
M  V30 7 2 7 8
M  V30 8 1 8 9
M  V30 9 2 9 10
M  V30 10 1 10 11
M  V30 11 1 11 12
M  V30 12 2 11 13
M  V30 13 1 13 14
M  V30 14 1 14 15
M  V30 15 1 15 16
M  V30 16 1 15 17
M  V30 17 1 15 18
M  V30 18 2 14 19
M  V30 19 1 19 20
M  V30 20 2 20 21
M  V30 21 1 21 2
M  V30 22 1 21 7
M  V30 23 1 19 9
M  V30 END BOND
M  V30 BEGIN COLLECTION
M  V30 MDLV30/STEREL1 ATOMS=(2 2 4)
M  V30 END COLLECTION
M  V30 END CTAB
M  END

$$$$
"""

    layers = hash_sdf(structure)
    expected_layers = {
      HashLayer.CANONICAL_SMILES:
      "C[C@H]1C[C@@H](C)c2cc3c(C(F)(F)F)cc(O)nc3cc2N1 |o1:1,3|",
      HashLayer.ESCAPE:
      "",
      HashLayer.FORMULA:
      "C15H15F3N2O",
      HashLayer.NO_STEREO_SMILES:
      "CC1CC(C)c2cc3c(C(F)(F)F)cc(O)nc3cc2N1",
      HashLayer.NO_STEREO_TAUTOMER_HASH:
      "CC1CC(C)[C]2[CH][C]3[C]([CH][C]2[N]1)[N][C]([O])[CH][C]3C(F)(F)F_2_0",
      HashLayer.SGROUP_DATA:
      "[]",
      HashLayer.TAUTOMER_HASH:
      "C[C@H]1C[C@@H](C)[C]2[CH][C]3[C]([CH][C]2[N]1)[N][C]([O])[CH][C]3C(F)(F)F_2_0 |o1:1,3|",
    }
    self.assertEqual(layers, expected_layers)

    # should have the same NO_STEREO_SMILES the structure with stereo removed
    mol = Chem.MolFromMolBlock(structure)
    Chem.rdmolops.RemoveStereochemistry(mol)
    stereo_insensitive_layers = hash_sdf(Chem.MolToMolBlock(mol))
    self.assertEqual(stereo_insensitive_layers[HashLayer.NO_STEREO_SMILES],
                     expected_layers[HashLayer.NO_STEREO_SMILES])

  def test_XBHEAD_XBCORR(self):
    # rdkit/Code/GraphMol/FileParsers/sgroup_test_data/repeat_groups_query1.mol
    structure = """
  Mrv1824 06192020192D

  0  0  0     0  0            999 V3000
M  V30 BEGIN CTAB
M  V30 COUNTS 6 6 1 0 0
M  V30 BEGIN ATOM
M  V30 1 C -1.25 0.7484 0 0
M  V30 2 C -2.5837 -0.0216 0 0
M  V30 3 C -2.5837 -1.5617 0 0
M  V30 4 C -1.25 -2.3317 0 0
M  V30 5 C 0.0837 -1.5617 0 0
M  V30 6 C 0.0837 -0.0216 0 0
M  V30 END ATOM
M  V30 BEGIN BOND
M  V30 1 1 1 2
M  V30 2 1 2 3
M  V30 3 1 3 4
M  V30 4 1 4 5
M  V30 5 1 5 6
M  V30 6 1 1 6
M  V30 END BOND
M  V30 BEGIN SGROUP
M  V30 1 SRU 0 ATOMS=(1 1) XBONDS=(2 1 6) XBHEAD=(2 6 1) XBCORR=(4 6 6 1 1) -
M  V30 BRKXYZ=(9 -2.02 -0.0216 0 -2.02 1.5184 0 0 0 0) BRKXYZ=(9 -0.48 1.5184 -
M  V30 0 -0.48 -0.0216 0 0 0 0) CONNECT=HT LABEL="1-3"
M  V30 END SGROUP
M  V30 END CTAB
M  END
"""

    expected_layers = {
      HashLayer.CANONICAL_SMILES: "C1CCCCC1",
      HashLayer.ESCAPE: "",
      HashLayer.FORMULA: "C6H12",
      HashLayer.NO_STEREO_SMILES: "C1CCCCC1",
      HashLayer.NO_STEREO_TAUTOMER_HASH: "C1CCCCC1_0_0",
      HashLayer.SGROUP_DATA:
      '[{"type": "SRU", "atoms": [2], "bonds": [[1, 2], [2, 3]], "index": 1, "connect": "HT", "label": "1-3", "XBHEAD": [[0, 5], [1, 0]], "XBCORR": [[[0, 5], [0, 5]], [[1, 0], [1, 0]]]}]',
      HashLayer.TAUTOMER_HASH: "C1CCCCC1_0_0",
    }

    layers = hash_sdf(structure)
    self.assertEqual(layers, expected_layers)

    # test for the same hash if the order in the XBHEAD block is changed
    structure.replace("XBHEAD=(2 6 1)", "XBHEAD=(2 1 6)")

    layers = hash_sdf(structure)
    self.assertEqual(layers, expected_layers)

  def test_data_sgroups(self):
    # see SHARED-7879
    structure = """
   Mrv1808 05312108392D

  0  0  0     0  0            999 V3000
M  V30 BEGIN CTAB
M  V30 COUNTS 7 6 1 0 0
M  V30 BEGIN ATOM
M  V30 1 C -5.6458 2.7017 0 0
M  V30 2 C -4.3121 1.9317 0 0
M  V30 3 C -2.9785 2.7017 0 0
M  V30 4 C -1.6448 1.9317 0 0
M  V30 5 C -0.3111 2.7017 0 0
M  V30 6 C 1.0226 1.9317 0 0
M  V30 7 C 2.3563 2.7017 0 0
M  V30 END ATOM
M  V30 BEGIN BOND
M  V30 1 1 1 2
M  V30 2 1 2 3
M  V30 3 1 3 4
M  V30 4 1 4 5
M  V30 5 1 5 6
M  V30 6 1 6 7
M  V30 END BOND
M  V30 BEGIN SGROUP
M  V30 1 DAT 0 ATOMS=(1 7) FIELDNAME="lambda max" FIELDINFO=nm -
M  V30 FIELDDISP="    1.4300   -0.7700    DRU   ALL  0       0" -
M  V30 MRV_FIELDDISP=0 FIELDDATA=250
M  V30 END SGROUP
M  V30 END CTAB
M  END
"""

    expected_layers = {
      HashLayer.CANONICAL_SMILES: 'CCCCCCC',
      HashLayer.ESCAPE: '',
      HashLayer.FORMULA: 'C7H16',
      HashLayer.NO_STEREO_SMILES: 'CCCCCCC',
      HashLayer.NO_STEREO_TAUTOMER_HASH: 'CCCCCCC_0_0',
      HashLayer.SGROUP_DATA: '[]',
      HashLayer.TAUTOMER_HASH: 'CCCCCCC_0_0'
    }

    expected_layers_v2 = {
      HashLayer.CANONICAL_SMILES: 'CCCCCCC',
      HashLayer.ESCAPE: '',
      HashLayer.FORMULA: 'C7H16',
      HashLayer.NO_STEREO_SMILES: 'CCCCCCC',
      HashLayer.NO_STEREO_TAUTOMER_HASH: '[CH3]-[CH2]-[CH2]-[CH2]-[CH2]-[CH2]-[CH3]_0_0',
      HashLayer.SGROUP_DATA: '[]',
      HashLayer.TAUTOMER_HASH: '[CH3]-[CH2]-[CH2]-[CH2]-[CH2]-[CH2]-[CH3]_0_0'
    }

    layers = hash_sdf(structure)
    self.assertEqual(layers, expected_layers)

    layers_v2 = hash_sdf(structure, enable_tautomer_hash_v2=True)
    self.assertEqual(layers_v2, expected_layers_v2)

  def test_stereo_imine_hash(self):
    stereo_mol = Chem.MolFromSmiles(r'[H]\N=C\C', sanitize=False)
    stereo = RegistrationHash.GetMolLayers(stereo_mol)

    no_stereo_mol = Chem.MolFromSmiles('N=CC', sanitize=False)
    no_stereo = RegistrationHash.GetMolLayers(no_stereo_mol)

    hydrogen_as_is_mol = Chem.MolFromSmiles('N=CC[H]', sanitize=False)
    hydrogen_as_is = RegistrationHash.GetMolLayers(hydrogen_as_is_mol)

    self.assertNotEqual(stereo[HashLayer.CANONICAL_SMILES], no_stereo[HashLayer.CANONICAL_SMILES])
    self.assertEqual(no_stereo[HashLayer.CANONICAL_SMILES],
                     hydrogen_as_is[HashLayer.CANONICAL_SMILES])

    self.assertEqual(stereo[HashLayer.NO_STEREO_SMILES], no_stereo[HashLayer.NO_STEREO_SMILES])
    self.assertEqual(no_stereo[HashLayer.NO_STEREO_SMILES],
                     hydrogen_as_is[HashLayer.NO_STEREO_SMILES])

    self.assertEqual(stereo[HashLayer.NO_STEREO_TAUTOMER_HASH],
                     no_stereo[HashLayer.NO_STEREO_TAUTOMER_HASH])
    self.assertEqual(no_stereo[HashLayer.NO_STEREO_TAUTOMER_HASH],
                     hydrogen_as_is[HashLayer.NO_STEREO_TAUTOMER_HASH])

    self.assertEqual(stereo[HashLayer.TAUTOMER_HASH], no_stereo[HashLayer.TAUTOMER_HASH])
    self.assertEqual(no_stereo[HashLayer.TAUTOMER_HASH], hydrogen_as_is[HashLayer.TAUTOMER_HASH])

  def test_empty(self):
    """Can the hasher operate on an empty molecule without traceback?"""
    mol = Chem.Mol()
    RegistrationHash.GetMolLayers(mol)

  def test_enhanced_stereo_canonicalizer(self):
    """Can we correctly canonicalize a molecule with enhanced stereo?"""
    groups = (
      # basics
      ('C[C@@H](O)[C@H](C)[C@@H](C)[C@@H](C)F |a:3,&1:1,7,&2:5,r|',
       'C[C@H](O)[C@H](C)[C@@H](C)[C@H](C)F |a:3,&1:1,7,&2:5,r|'),

      # renumbering the groups
      ('C[C@@H](O)[C@H](C)[C@@H](C)[C@@H](C)F |a:3,&1:1,7,&2:5,r|',
       'C[C@@H](O)[C@H](C)[C@@H](C)[C@@H](C)F |a:3,&2:1,7,&1:5,r|'),

      # AND/OR canonicalization
      ('C[C@@H](O)[C@H](C)[C@@H](C)[C@@H](C)O |&3:3,o1:7,&2:1,&3:5,r|',
       'C[C@@H](O)[C@H](C)[C@@H](C)[C@@H](C)O |&3:3,&2:7,o1:1,&3:5,r|'),

      # Dan's example in SHARED-7811 (corrected)
      (
        r'C[C@@H]1CC(C[C@@H](C)[C@@H](C)O)C[C@H](C)C1 |o1:1,7,o2:5,11|',  # s1
        r'C[C@H](O)[C@H](C)CC1C[C@H](C)C[C@@H](C)C1 |o1:1,8;o2:3,11|',  # s2
        r'C[C@H]1CC(C[C@@H](C)[C@@H](C)O)C[C@@H](C)C1 |o1:1,5,o2:7,11|'  # s3 fixed!
        r'C[C@@H](O)[C@H](C)CC1C[C@H](C)C[C@H](C)C1 |o1:1,11;o2:3,8|',  # s4
      ),
    )
    for mols_smis in groups:
      csmis = set()
      for smi in mols_smis:
        mol = Chem.MolFromSmiles(smi)
        csmi = Chem.MolToCXSmiles(mol)
        self.assertIsNotNone(csmi)
        csmis.add(csmi)
      self.assertEqual(len(csmis), 1)

  def test_enhanced_stereo_canonicalizer_non_matching(self):
    # Uncorrected example in SHARED-7811. These are NOT equivalent,
    # since the groups are defined on non-interconvertible pairs of atomss
    s1 = r'C[C@@H]1CC(C[C@@H](C)[C@@H](C)O)C[C@H](C)C1 |o1:1,7,o2:5,11|'
    s3 = r'C[C@H]1CC(C[C@@H](C)[C@@H](C)O)C[C@@H](C)C1 |o1:1,11,o2:5,7|'

    csmis = set()
    for smi in (s1, s3):
      mol = Chem.MolFromSmiles(smi)
      csmi = Chem.MolToCXSmiles(mol)
      csmis.add(csmi)
    self.assertEqual(len(csmis), 2)

  def test_enhanced_stereo_regex(self):
    cxsmileses = ((
      'C[C@@H](O)[C@H](C)[C@@H](C)[C@@H](C)F |a:3,&1:1,7,&2:5,r|',
      'C[C@@H](O)[C@H](C)[C@@H](C)[C@@H](C)O |&3:3,o1:7,&2:1,&3:5,r|',
      'C[C@@H](O)[C@H](C)CC1C[C@H](C)C[C@H](C)C1 |o1:1,11;o2:3,8|',
    ))
    for cxsmiles in cxsmileses:
      self.assertEqual(cxsmiles.count(':'),
                       len(RegistrationHash.ENHANCED_STEREO_GROUP_REGEX.findall(cxsmiles)))

  def test_shared_7984(self):
    smi = '[2H]C([2H])=C([2H])C(=O)O'
    no_stereo_mol = Chem.MolFromSmiles(smi)
    self.assertEqual(no_stereo_mol.GetNumAtoms(), 8)

    stereo_mol = Chem.Mol(no_stereo_mol)

    # We consider a double bond with these features equal to a STEREONONE one.
    stereo_mol.GetBondWithIdx(2).SetStereo(Chem.BondStereo.STEREOANY)
    stereo_mol.GetBondWithIdx(2).SetBondDir(Chem.BondDir.EITHERDOUBLE)

    no_stereo_hash = RegistrationHash.GetMolLayers(no_stereo_mol)
    stereo_hash = RegistrationHash.GetMolLayers(stereo_mol)

    self.assertEqual(no_stereo_hash, stereo_hash)

  def test_non_matching_pseudoatom_label(self):

    pol_sdf = """
     RDKit          2D

  3  2  0  0  0  0  0  0  0  0999 V2000
    2.5981   -0.0000    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    1.2990    0.7500    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    0.0000    0.0000    0.0000 Pol 0  0  0  0  0  0  0  0  0  0  0  0
  1  2  1  0
  2  3  1  0
M  END"""

    mod_sdf = """
     RDKit          2D

  3  2  0  0  0  0  0  0  0  0999 V2000
    2.5981   -0.0000    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    1.2990    0.7500    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    0.0000    0.0000    0.0000 Mod 0  0  0  0  0  0  0  0  0  0  0  0
  1  2  1  0
  2  3  1  0
M  END"""

    carbon_sdf = """
     RDKit          2D

  3  2  0  0  0  0  0  0  0  0999 V2000
    2.5981   -0.0000    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    1.2990    0.7500    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    0.0000    0.0000    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
  1  2  1  0
  2  3  1  0
M  END"""

    pol_hash = RegistrationHash.GetMolHash(hash_sdf(pol_sdf))
    mod_hash = RegistrationHash.GetMolHash(hash_sdf(mod_sdf))
    carbon_hash = RegistrationHash.GetMolHash(hash_sdf(carbon_sdf))

    hashes = {pol_hash, mod_hash, carbon_hash}
    self.assertEqual(len(hashes), 3)

  def test_matching_pseudoatom_label(self):

    pol_sdf = """
     RDKit          2D

  3  2  0  0  0  0  0  0  0  0999 V2000
    2.5981   -0.0000    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    1.2990    0.7500    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    0.0000    0.0000    0.0000 Pol 0  0  0  0  0  0  0  0  0  0  0  0
  1  2  1  0
  2  3  1  0
M  END"""

    pol_sdf_2 = """
     RDKit          2D

  3  2  0  0  0  0  0  0  0  0999 V2000
    2.5981   -0.0000    0.0000 Pol 0  0  0  0  0  0  0  0  0  0  0  0
    1.2990    0.7500    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    0.0000    0.0000    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
  1  2  1  0
  2  3  1  0
M  END"""

    pol_hash = RegistrationHash.GetMolHash(hash_sdf(pol_sdf))
    pol_hash_2 = RegistrationHash.GetMolHash(hash_sdf(pol_sdf_2))

    hashes = {pol_hash, pol_hash_2}
    self.assertEqual(len(hashes), 1)

  def test_hash_schemes(self):

    def get_hash(m, scheme):
      hash = RegistrationHash.GetMolHash(RegistrationHash.GetMolLayers(m), scheme)
      self.assertEqual(len(hash), 40)  # SHA-1 hashes are 40 hex digits long
      return hash

    def same_hash(m1, m2, scheme):
      return get_hash(m1, scheme) == get_hash(m2, scheme)

    mol1 = Chem.MolFromSmiles("C[C@H]1CC(C[C@@H](C)[C@@H](C)O)C[C@@H](C)C1 |o1:1,11,o2:5,7|")
    self.assertEqual(mol1.GetNumAtoms(), 14)

    all_layers_hash = get_hash(mol1, RegistrationHash.HashScheme.ALL_LAYERS)
    stereo_insensitive_layers_hash = get_hash(mol1,
                                              RegistrationHash.HashScheme.STEREO_INSENSITIVE_LAYERS)
    tautomer_insensitive_layers_hash = get_hash(
      mol1, RegistrationHash.HashScheme.TAUTOMER_INSENSITIVE_LAYERS)

    self.assertNotEqual(all_layers_hash, tautomer_insensitive_layers_hash)
    self.assertNotEqual(all_layers_hash, stereo_insensitive_layers_hash)
    self.assertNotEqual(stereo_insensitive_layers_hash, tautomer_insensitive_layers_hash)

    # Compare stereoisomers
    mol2 = Chem.MolFromSmiles("CC(O)C(C)CC1CC(C)CC(C)C1")
    self.assertFalse(same_hash(mol1, mol2, RegistrationHash.HashScheme.ALL_LAYERS))
    self.assertTrue(same_hash(mol1, mol2, RegistrationHash.HashScheme.STEREO_INSENSITIVE_LAYERS))
    self.assertFalse(same_hash(mol1, mol2, RegistrationHash.HashScheme.TAUTOMER_INSENSITIVE_LAYERS))

    # Compare tautomers
    mol1 = Chem.MolFromSmiles("N1C=NC2=CN=CN=C12")
    mol2 = Chem.MolFromSmiles("N1C=NC2=NC=NC=C12")
    self.assertFalse(same_hash(mol1, mol2, RegistrationHash.HashScheme.ALL_LAYERS))
    self.assertFalse(same_hash(mol1, mol2, RegistrationHash.HashScheme.STEREO_INSENSITIVE_LAYERS))
    self.assertTrue(same_hash(mol1, mol2, RegistrationHash.HashScheme.TAUTOMER_INSENSITIVE_LAYERS))

    # Compare with and without sgroup data we canonicalize on (DAT/SRU/COP)
    smiles = "CNCC(=O)OC"
    mol1 = Chem.MolFromSmiles(smiles)
    mol2 = Chem.MolFromSmiles(f"{smiles} |Sg:n:5,3,4::ht|")
    # ...where all hash schemes account for sgroup data
    self.assertFalse(same_hash(mol1, mol2, RegistrationHash.HashScheme.ALL_LAYERS))
    self.assertFalse(same_hash(mol1, mol2, RegistrationHash.HashScheme.STEREO_INSENSITIVE_LAYERS))
    self.assertFalse(same_hash(mol1, mol2, RegistrationHash.HashScheme.TAUTOMER_INSENSITIVE_LAYERS))

  def test_no_stereo_hash(self):
    """
        Check that none of the layers involved in the STEREO_INSENSITIVE_LAYERS HashScheme
        includes any SMILES stereo markers
        """
    smi = r'C\C=C/[C@@H](C)Cl'
    mol = Chem.MolFromSmiles(smi)
    self.assertEqual(mol.GetNumAtoms(), 6)

    layers = RegistrationHash.GetMolLayers(mol)

    for layer in RegistrationHash.HashScheme.STEREO_INSENSITIVE_LAYERS.value:
      self.assertNotIn('@', layers[layer])
      self.assertNotIn('/', layers[layer])
      self.assertNotIn('\\', layers[layer])

  def test_atom_map_numbers(self):
    """
        Check that atom map numbers are stripped in the molhash (SHARED-8007)
        """
    mol_w_atom_map = Chem.MolFromSmiles('Oc1cc([*:1])ccn1')
    mol = Chem.MolFromSmiles('Oc1cc([*])ccn1')
    atom_map_hash_layers = RegistrationHash.GetMolLayers(mol_w_atom_map)
    hash_layers = RegistrationHash.GetMolLayers(mol)
    assert atom_map_hash_layers[HashLayer.CANONICAL_SMILES] == hash_layers[
      HashLayer.CANONICAL_SMILES]

  def test_compare_stereoisomers(self):
    stereo_mol = Chem.MolFromSmiles("C[C@H]1CC(C[C@@H](C)[C@@H](C)O)C[C@@H](C)C1 |o1:1,11,o2:5,7|")
    no_stereo_mol = Chem.MolFromSmiles("CC(O)C(C)CC1CC(C)CC(C)C1")

    stereo_layers = RegistrationHash.GetMolLayers(stereo_mol)
    stereo_insensitive_layers = RegistrationHash.GetMolLayers(no_stereo_mol)
    self.assertEqual(
      RegistrationHash.GetMolHash(stereo_layers,
                                  RegistrationHash.HashScheme.STEREO_INSENSITIVE_LAYERS),
      RegistrationHash.GetMolHash(stereo_insensitive_layers,
                                  RegistrationHash.HashScheme.STEREO_INSENSITIVE_LAYERS))
    self.assertEqual(stereo_layers[HashLayer.NO_STEREO_SMILES], "CC1CC(C)CC(CC(C)C(C)O)C1")
    self.assertEqual(stereo_insensitive_layers[HashLayer.NO_STEREO_SMILES],
                     "CC1CC(C)CC(CC(C)C(C)O)C1")

  def _compare_layers_with_expected_diffs(self, new_layers, old_layers, expected_diffs):
    for k in old_layers:
      if k not in expected_diffs:
        self.assertEqual(new_layers[k], old_layers[k])
      else:
        self.assertEqual(new_layers[k], expected_diffs[k])

  def test_sgroup_data_hashes(self):
    structure1 = '''
  Mrv2108 11252113552D

  0  0  0     0  0            999 V3000
M  V30 BEGIN CTAB
M  V30 COUNTS 30 34 2 0 0
M  V30 BEGIN ATOM
M  V30 1 C 8.78 -9.2456 0 0
M  V30 2 C 7.2925 -8.8469 0 0
M  V30 3 C 6.2033 -9.9358 0 0
M  V30 4 C 6.6021 -11.4235 0 0
M  V30 5 C 8.0896 -11.8222 0 0
M  V30 6 C 9.1786 -10.733 0 0
M  V30 7 C 10.6662 -11.1317 0 0
M  V30 8 C 11.7552 -10.0426 0 0
M  V30 9 C 11.3565 -8.5552 0 0
M  V30 10 C 9.8688 -8.1565 0 0
M  V30 11 C 13.2428 -10.441 0 0
M  V30 12 C 14.3317 -9.3519 0 0 CFG=2
M  V30 13 C 15.8524 -9.5927 0 0
M  V30 14 C 16.5517 -8.2207 0 0
M  V30 15 C 15.463 -7.1316 0 0
M  V30 16 C 14.0905 -7.8307 0 0
M  V30 17 C 16.5515 -10.9649 0 0
M  V30 18 C 8.4882 -13.3096 0 0
M  V30 19 C 5.5131 -12.5124 0 0
M  V30 20 C 5.9117 -14.0001 0 0
M  V30 21 C 4.8227 -15.0888 0 0
M  V30 22 C 3.3352 -14.6902 0 0
M  V30 23 C 2.9365 -13.2027 0 0
M  V30 24 C 4.0255 -12.1139 0 0
M  V30 25 C 3.6269 -10.6262 0 0
M  V30 26 Cl 4.7157 -9.5372 0 0
M  V30 27 C 5.2472 -16.5691 0 0
M  V30 28 C 6.7414 -16.9417 0 0
M  V30 29 C 7.8112 -15.8339 0 0
M  V30 30 C 7.3867 -14.3535 0 0
M  V30 END ATOM
M  V30 BEGIN BOND
M  V30 1 2 1 2
M  V30 2 1 2 3
M  V30 3 2 3 4
M  V30 4 1 4 5
M  V30 5 2 5 6
M  V30 6 1 6 1
M  V30 7 1 6 7
M  V30 8 2 7 8
M  V30 9 1 8 9
M  V30 10 2 9 10
M  V30 11 1 10 1
M  V30 12 1 12 11 CFG=1
M  V30 13 1 12 13
M  V30 14 1 13 14
M  V30 15 1 14 15
M  V30 16 1 15 16
M  V30 17 1 16 12
M  V30 18 1 13 17
M  V30 19 1 5 18
M  V30 20 1 4 19
M  V30 21 1 19 20
M  V30 22 2 20 21
M  V30 23 1 21 22
M  V30 24 2 22 23
M  V30 25 1 23 24
M  V30 26 2 24 19
M  V30 27 1 24 25
M  V30 28 1 3 26
M  V30 29 1 21 27
M  V30 30 1 20 30
M  V30 31 2 29 30
M  V30 32 1 28 29
M  V30 33 2 27 28
M  V30 34 1 8 11
M  V30 END BOND
M  V30 BEGIN SGROUP
M  V30 1 DAT 0 ATOMS=(1 12) FIELDNAME=stereolabel -
M  V30 FIELDDISP="   14.8391  -10.2748    DA    ALL  1       5" FIELDDATA=R
M  V30 2 DAT 0 ATOMS=(2 4 19) FIELDNAME=STEREOLABEL -
M  V30 FIELDDISP="    6.4455  -12.6102    DA    ALL  1       5" FIELDDATA=R
M  V30 END SGROUP
M  V30 END CTAB
M  END
'''
    structure2 = structure1.replace('FIELDDATA=R', 'FIELDDATA=S')
    structure3 = structure1.replace('FIELDDATA=R', 'FIELDDATA=S', 1)

    # default behavior is that these are all the same:
    layers1 = hash_sdf(structure1)
    layers2 = hash_sdf(structure2)
    expected_layers1 = {
      HashLayer.CANONICAL_SMILES:
      "Cc1ccc2ccccc2c1-c1c(Cl)cc2ccc(C[C@H]3CCCC3C)cc2c1C",
      HashLayer.ESCAPE:
      "",
      HashLayer.FORMULA:
      "C29H29Cl",
      HashLayer.NO_STEREO_SMILES:
      "Cc1ccc2ccccc2c1-c1c(Cl)cc2ccc(CC3CCCC3C)cc2c1C",
      HashLayer.NO_STEREO_TAUTOMER_HASH:
      "C[C]1[CH][CH][C]2[CH][CH][CH][CH][C]2[C]1[C]1[C](Cl)[CH][C]2[CH][CH][C](CC3CCCC3C)[CH][C]2[C]1C_0_0",
      HashLayer.SGROUP_DATA:
      "[]",
      HashLayer.TAUTOMER_HASH:
      "C[C]1[CH][CH][C]2[CH][CH][CH][CH][C]2[C]1[C]1[C](Cl)[CH][C]2[CH][CH][C](C[C@H]3CCCC3C)[CH][C]2[C]1C_0_0",
    }
    self.assertEqual(layers1, expected_layers1)
    self.assertEqual(layers2, layers1)

    # using both stereo labels they are all different from each other:
    layers1a = hash_sdf(structure1, data_field_names=['STEREOLABEL', 'stereolabel'])
    expected_diffs = {
      HashLayer.SGROUP_DATA:
      '[{"fieldName": "STEREOLABEL", "atom": [10, 11], "bonds": [], "value": "R"}, {"fieldName": "stereolabel", "atom": [20], "bonds": [], "value": "R"}]',
    }
    self._compare_layers_with_expected_diffs(layers1a, layers1, expected_diffs)

    layers2a = hash_sdf(structure2, data_field_names=['STEREOLABEL', 'stereolabel'])
    expected_diffs = {
      HashLayer.SGROUP_DATA:
      '[{"fieldName": "STEREOLABEL", "atom": [10, 11], "bonds": [], "value": "S"}, {"fieldName": "stereolabel", "atom": [20], "bonds": [], "value": "S"}]',
    }
    self._compare_layers_with_expected_diffs(layers2a, layers1, expected_diffs)

    layers3a = hash_sdf(structure3, data_field_names=['STEREOLABEL', 'stereolabel'])
    expected_diffs = {
      HashLayer.SGROUP_DATA:
      '[{"fieldName": "STEREOLABEL", "atom": [10, 11], "bonds": [], "value": "R"}, {"fieldName": "stereolabel", "atom": [20], "bonds": [], "value": "S"}]',
    }
    self._compare_layers_with_expected_diffs(layers3a, layers1, expected_diffs)

    # ensure that we can pick the data field to use:
    layers1b = hash_sdf(structure1, data_field_names=['stereolabel'])
    expected_diffs = {
      HashLayer.SGROUP_DATA:
      '[{"fieldName": "stereolabel", "atom": [20], "bonds": [], "value": "R"}]',
    }
    self._compare_layers_with_expected_diffs(layers1b, layers1, expected_diffs)

  def test_escape_layer(self):
    mol = Chem.MolFromSmiles("c1ccccc1")

    layers = RegistrationHash.GetMolLayers(mol)
    layers_with_escape = RegistrationHash.GetMolLayers(mol, escape="make it unique")
    assert layers_with_escape[HashLayer.ESCAPE] == "make it unique"
    for k in layers_with_escape:
      if k != HashLayer.ESCAPE:
        assert layers_with_escape[k] == layers[k]

  def test_shared_8516(self):
    no_explicit_h_mol = Chem.MolFromSmiles('C[C@H](N)C(=O)O', sanitize=False)
    with_explicit_h_mol = Chem.MolFromSmiles('[H][C@@](C)(N)C(=O)O', sanitize=False)
    no_explicit_h_layers = RegistrationHash.GetMolLayers(no_explicit_h_mol)
    with_explicit_h_layers = RegistrationHash.GetMolLayers(with_explicit_h_mol)
    self.assertEqual(no_explicit_h_layers, with_explicit_h_layers)

  def test_mol_layers_distinguish_carbon_isotopes(self):
    benzene = Chem.MolFromSmiles("c1ccccc1")
    benzene_with_c13 = Chem.MolFromSmiles("[13c]1ccccc1")

    layers1 = RegistrationHash.GetMolLayers(benzene)
    layers2 = RegistrationHash.GetMolLayers(benzene_with_c13)

    self.assertNotEqual(layers1, layers2)

  def test_mol_layers_distinguish_non_tautomeric_hydrogen_isotopes(self):
    benzene_with_deuterium = Chem.MolFromSmiles("[2H]c1ccccc1")
    layers = RegistrationHash.GetMolLayers(benzene_with_deuterium)

    expected_layers = {
      HashLayer.CANONICAL_SMILES: "[2H]c1ccccc1",
      HashLayer.ESCAPE: "",
      HashLayer.FORMULA: "C6H6",
      HashLayer.NO_STEREO_SMILES: "[2H]c1ccccc1",
      HashLayer.NO_STEREO_TAUTOMER_HASH: "[2H][C]1[CH][CH][CH][CH][CH]1_0_0",
      HashLayer.SGROUP_DATA: "[]",
      HashLayer.TAUTOMER_HASH: "[2H][C]1[CH][CH][CH][CH][CH]1_0_0"
    }

    assert layers == expected_layers

  def test_mol_layers_distinguish_tautomeric_hydrogen_isotopes(self):
    mol1 = Chem.MolFromSmiles("[2H]N1C=NC2=CN=CN=C12")
    mol2 = Chem.MolFromSmiles("[2H]N1C=NC2=NC=NC=C12")

    layers1 = RegistrationHash.GetMolLayers(mol1)
    layers2 = RegistrationHash.GetMolLayers(mol2)

    expected_layers1 = {
      HashLayer.CANONICAL_SMILES: "[2H]n1cnc2cncnc21",
      HashLayer.ESCAPE: "",
      HashLayer.FORMULA: "C5H4N4",
      HashLayer.NO_STEREO_SMILES: "[2H]n1cnc2cncnc21",
      HashLayer.NO_STEREO_TAUTOMER_HASH: "[2H]N1[CH][N][C]2[CH][N][CH][N][C]21_0_0",
      HashLayer.SGROUP_DATA: "[]",
      HashLayer.TAUTOMER_HASH: "[2H]N1[CH][N][C]2[CH][N][CH][N][C]21_0_0"
    }

    expected_layers2 = {
      HashLayer.CANONICAL_SMILES: "[2H]n1cnc2ncncc21",
      HashLayer.ESCAPE: "",
      HashLayer.FORMULA: "C5H4N4",
      HashLayer.NO_STEREO_SMILES: "[2H]n1cnc2ncncc21",
      HashLayer.NO_STEREO_TAUTOMER_HASH: "[2H]N1[CH][N][C]2[N][CH][N][CH][C]21_0_0",
      HashLayer.SGROUP_DATA: "[]",
      HashLayer.TAUTOMER_HASH: "[2H]N1[CH][N][C]2[N][CH][N][CH][C]21_0_0"
    }

    self.assertEqual(layers1, expected_layers1)
    self.assertEqual(layers2, expected_layers2)

  def testIodine(self):
    """Does stereo group canonicalization mess up isotopes?"""
    mol = Chem.MolFromSmiles('CC[C@H](C)[999C@H](C)O  |o1:2,4|')
    layers = RegistrationHash.GetMolLayers(mol)
    self.assertEqual(layers[HashLayer.CANONICAL_SMILES], 'CC[C@H](C)[999C@H](C)O |o1:2,4|')

  def testBadBondDir(self):
    """"""
    molBlock = """
     RDKit          2D

  0  0  0  0  0  0  0  0  0  0999 V3000
M  V30 BEGIN CTAB
M  V30 COUNTS 12 12 0 0 0
M  V30 BEGIN ATOM
M  V30 1 C 5.192672 -7.044349 0.000000 0
M  V30 2 C 5.272538 -5.675621 0.000000 0
M  V30 3 C 6.415595 -5.133214 0.000000 0
M  V30 4 C 7.487292 -5.854749 0.000000 0
M  V30 5 C 6.572156 -3.950482 0.000000 0
M  V30 6 C 5.935220 -2.698719 0.000000 0
M  V30 7 C 4.531603 -2.325867 0.000000 0
M  V30 8 C 4.159340 -0.962544 0.000000 0
M  V30 9 C 2.686034 -0.657857 0.000000 0
M  V30 10 C 5.166201 0.053779 0.000000 0
M  V30 11 C 6.621450 -0.316643 0.000000 0
M  V30 12 C 6.973858 -1.673599 0.000000 0
M  V30 END ATOM
M  V30 BEGIN BOND
M  V30 1 1 1 2
M  V30 2 1 2 3
M  V30 3 1 3 4
M  V30 4 2 3 5
M  V30 5 1 6 5 CFG=3
M  V30 6 1 6 7
M  V30 7 1 7 8
M  V30 8 2 8 9
M  V30 9 1 8 10
M  V30 10 1 11 10
M  V30 11 1 6 12
M  V30 12 1 11 12
M  V30 END BOND
M  V30 BEGIN COLLECTION
M  V30 MDLV30/STERAC1 ATOMS=(1 6)
M  V30 END COLLECTION
M  V30 END CTAB
M  END
$$$$
"""
    mol = Chem.MolFromMolBlock(molBlock)
    # Simulate reapplying the wedging from the original molBlock
    mol.GetBondBetweenAtoms(4, 5).SetBondDir(Chem.BondDir.BEGINWEDGE)
    # this shouldn't throw an exception
    RegistrationHash.GetMolLayers(mol)

  def test_tautomer_v2_hash(self):
    """
        Check that using v2 of the tautomer hash works
    """
    enol = Chem.MolFromSmiles('CC=CO')
    keto = Chem.MolFromSmiles('CCC=O')

    # Default, v1 of the tautomer hash:
    enol_layers = RegistrationHash.GetMolLayers(enol)
    keto_layers = RegistrationHash.GetMolLayers(keto)

    for layer in (RegistrationHash.HashLayer.TAUTOMER_HASH,
                  RegistrationHash.HashLayer.NO_STEREO_TAUTOMER_HASH):
      self.assertNotEqual(enol_layers[layer], keto_layers[layer])

    self.assertNotEqual(
      RegistrationHash.GetMolHash(
        enol_layers, hash_scheme=RegistrationHash.HashScheme.TAUTOMER_INSENSITIVE_LAYERS),
      RegistrationHash.GetMolHash(
        keto_layers, hash_scheme=RegistrationHash.HashScheme.TAUTOMER_INSENSITIVE_LAYERS))

    # v2 of the tautomer hash:
    enol_layers = RegistrationHash.GetMolLayers(enol, enable_tautomer_hash_v2=True)
    self.assertIn(RegistrationHash.HashLayer.TAUTOMER_HASH, enol_layers)

    keto_layers = RegistrationHash.GetMolLayers(keto, enable_tautomer_hash_v2=True)

    for layer in (RegistrationHash.HashLayer.TAUTOMER_HASH,
                  RegistrationHash.HashLayer.NO_STEREO_TAUTOMER_HASH):
      self.assertEqual(enol_layers[layer], keto_layers[layer])

    self.assertEqual(
      RegistrationHash.GetMolHash(
        enol_layers, hash_scheme=RegistrationHash.HashScheme.TAUTOMER_INSENSITIVE_LAYERS),
      RegistrationHash.GetMolHash(
        keto_layers, hash_scheme=RegistrationHash.HashScheme.TAUTOMER_INSENSITIVE_LAYERS))

  def testAtropisomer(self):
    atropTestFile = os.path.join(
      RDConfig.RDBaseDir, 'Code/GraphMol/FileParsers/test_data/atropisomers/RP-6306_atrop1.sdf')
    mol = Chem.MolFromMolFile(atropTestFile)

    bond = mol.GetBondWithIdx(3)
    self.assertEqual(bond.GetStereo(), Chem.BondStereo.STEREOATROPCW)

    layersCw = RegistrationHash.GetMolLayers(mol)

    smiles1, smiExt1 = layersCw[HashLayer.CANONICAL_SMILES].split()
    taut1, tautExt1 = layersCw[HashLayer.TAUTOMER_HASH].split()
    self.assertEqual(smiles1, 'Cc1cc2c(C(N)=O)c(N)n(-c3c(C)ccc(O)c3C)c2nc1C')
    self.assertEqual(
      taut1,
      'C[C]1[CH][C]2[C]([C]([N])[O])[C]([N])N([C]3[C](C)[CH][CH][C]([O])[C]3C)[C]2[N][C]1C_5_0')
    self.assertEqual(smiExt1, '|wD:10.9|')
    self.assertEqual(smiExt1, tautExt1)

    # Now look at the other atropisomer
    bond = mol.GetBondWithIdx(3)
    bond.SetStereo(Chem.BondStereo.STEREOATROPCCW)

    # Hashes should not match
    layersCcw = RegistrationHash.GetMolLayers(mol)
    self.assertNotEqual(layersCw, layersCcw)

    smiles2, smiExt2 = layersCcw[HashLayer.CANONICAL_SMILES].split()
    taut2, tautExt2 = layersCcw[HashLayer.TAUTOMER_HASH].split()

    # SMILES and tautomer hash should be the same, but the extensions must be different
    self.assertEqual(smiles2, smiles1)
    self.assertEqual(taut2, taut1)
    self.assertEqual(smiExt2, '|wD:10.20|')  # same atom, but "down" wedge on a different bond
    self.assertEqual(smiExt2, tautExt2)


if __name__ == '__main__':  # pragma: nocover
  unittest.main()



# $Id$
#
# Copyright (C) 2006-2008 Greg Landrum
#
#  @@ All Rights Reserved @@
#

import io
import os
import pickle
import sys
import unittest

from rdkit import Chem, RDConfig
from rdkit.Chem import ChemicalFeatures
from rdkit.Geometry import rdGeometry as geom


def feq(v1, v2, tol2=1e-4):
  return abs(v1 - v2) <= tol2


def lstFeq(l1, l2, tol=1.e-4):
  if len(l1) != len(l2):
    return 0
  for ll1, ll2 in zip(l1, l2):
    if not feq(ll1, ll2, tol):
      return 0
  return 1


def ptFeq(pt1, pt2, tol=0.0001):
  dist = pt1.Distance(pt2)
  return feq(dist, 0.0, tol)


class TestCase(unittest.TestCase):

  def setUp(self):
    pass

  def testBasic(self):
    ffeat = ChemicalFeatures.FreeChemicalFeature()
    ffeat.SetId(123)
    pos = ffeat.GetId()
    self.assertTrue(pos == 123)
    ffeat.SetFamily("HBondDonor")
    self.assertTrue(ffeat.GetFamily() == "HBondDonor")
    ffeat.SetPos(geom.Point3D(1.0, 2.0, 3.0))
    pos = ffeat.GetPos()
    self.assertTrue(ptFeq(pos, geom.Point3D(1.0, 2.0, 3.0)))
    ffeat.SetType("HBondDonor1")
    self.assertTrue(ffeat.GetType() == "HBondDonor1")

    ffeat = ChemicalFeatures.FreeChemicalFeature("HBondDonor", "HBondDonor1",
                                                 geom.Point3D(1.0, 2.0, 3.0))
    self.assertTrue(ffeat.GetId() == -1)
    self.assertTrue(ffeat.GetFamily() == "HBondDonor")
    self.assertTrue(ffeat.GetType() == "HBondDonor1")

    ffeat = ChemicalFeatures.FreeChemicalFeature("HBondDonor", "HBondDonor1",
                                                 geom.Point3D(1.0, 2.0, 3.0), id=123)
    self.assertTrue(ffeat.GetId() == 123)
    self.assertTrue(ffeat.GetFamily() == "HBondDonor")
    self.assertTrue(ffeat.GetType() == "HBondDonor1")

    pos = ffeat.GetPos()
    self.assertTrue(ptFeq(pos, geom.Point3D(1.0, 2.0, 3.0)))

    ffeat = ChemicalFeatures.FreeChemicalFeature(id=123, type="HBondDonor1", family="HBondDonor",
                                                 loc=geom.Point3D(1.0, 2.0, 3.0))
    self.assertTrue(ffeat.GetId() == 123)
    self.assertTrue(ffeat.GetFamily() == "HBondDonor")
    self.assertTrue(ffeat.GetType() == "HBondDonor1")
    pos = ffeat.GetPos()
    self.assertTrue(ptFeq(pos, geom.Point3D(1.0, 2.0, 3.0)))

  def testPickle(self):
    ffeat = ChemicalFeatures.FreeChemicalFeature("HBondDonor", "HBondDonor1",
                                                 geom.Point3D(1.0, 2.0, 3.0), 123)
    pkl = pickle.dumps(ffeat)
    ffeat2 = pickle.loads(pkl, encoding='bytes')
    self.assertTrue(ffeat2.GetId() == ffeat.GetId())
    self.assertTrue(ffeat2.GetFamily() == ffeat.GetFamily())
    self.assertTrue(ffeat2.GetType() == ffeat.GetType())
    self.assertTrue(ptFeq(ffeat2.GetPos(), ffeat.GetPos()))

    # Check that the old pickled versions have not been broken
    inTF = open(os.path.join(RDConfig.RDBaseDir, 'Code/ChemicalFeatures/Wrap/testData/feat.pkl'),
                'r')
    buf = inTF.read().replace('\r\n', '\n').encode('utf-8')
    inTF.close()
    inF = io.BytesIO(buf)
    ffeat2 = pickle.load(inF, encoding='bytes')
    # this version (1.0) does not have an id in the byte stream
    self.assertTrue(ffeat2.GetFamily() == ffeat.GetFamily())
    self.assertTrue(ffeat2.GetType() == ffeat.GetType())
    self.assertTrue(ptFeq(ffeat2.GetPos(), ffeat.GetPos()))

    # Test the new version also has the id and works as expected

    # uncomment the following to generate (overwrite) new version of pickled
    # data file
    #pickle.dump(ffeat,file(os.path.join(RDConfig.RDBaseDir, 'Code/ChemicalFeatures/Wrap/testData/featv2.pkl'),'wb+'))
    inTF = open(os.path.join(RDConfig.RDBaseDir, 'Code/ChemicalFeatures/Wrap/testData/featv2.pkl'),
                'r')
    buf = inTF.read().replace('\r\n', '\n').encode('utf-8')
    inTF.close()
    inF = io.BytesIO(buf)
    ffeat2 = pickle.load(inF, encoding='bytes')
    self.assertTrue(ffeat2.GetId() == ffeat.GetId())
    self.assertTrue(ffeat2.GetFamily() == ffeat.GetFamily())
    self.assertTrue(ffeat2.GetType() == ffeat.GetType())
    self.assertTrue(ptFeq(ffeat2.GetPos(), ffeat.GetPos()))


if __name__ == '__main__':
  print("Testing ChemicalFeatures Wrapper code:")
  unittest.main()



import gzip

from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem.PyMol import MolViewer

try:
  v = MolViewer()
  v.DeleteAll()
except Exception:
  v = None

outf = open('chembl_20_chiral.problems.smi', 'w+')
noconff = open('chembl_20_chiral.noconfs.smi', 'w+')
for i, line in enumerate(gzip.open('../Data/chembl_20_chiral.smi.gz')):
  line = line.strip().decode().split(' ')
  mol = Chem.MolFromSmiles(line[0])
  if not mol:
    continue
  cents = Chem.FindMolChiralCenters(mol, includeUnassigned=True)
  if len([y for x, y in cents if y == '?']):
    continue
  nm = line[1]
  csmi = Chem.MolToSmiles(mol, True)
  for j in range(100):
    mh = Chem.AddHs(mol)
    ok = AllChem.EmbedMolecule(mh, randomSeed=j + 1)
    if ok >= 0:
      Chem.AssignAtomChiralTagsFromStructure(mh)
      newm = Chem.RemoveHs(mh)
      smi = Chem.MolToSmiles(newm, True)
      if smi != csmi:
        print('%d %d %s:\n%s\n%s' % (i, j, nm, csmi, smi))
        print('%s %s %d' % (line[0], line[1], j + 1), file=outf)

        if v is not None:
          v.ShowMol(mh, name='%s-%d' % (nm, j), showOnly=False)
          break  # move immediately onto the next molecule
    else:
      print('noconf %d %d %s: %s' % (i, j, nm, line[0]))
      print('%s %s %d' % (line[0], line[1], j + 1), file=noconff)

  print('Done with mol %d' % i)



# $Id$
#
#  Copyright (C) 2007  greg Landrum
#
#   @@ All Rights Reserved @@
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#

import doctest
import unittest
from io import StringIO

from rdkit.VLib import Filter, Node, Output, Supply, Transform


def load_tests(loader, tests, ignore):
  """ Add the Doctests from the module """
  tests.addTests(doctest.DocTestSuite(Filter, optionflags=doctest.ELLIPSIS))
  tests.addTests(doctest.DocTestSuite(Node, optionflags=doctest.ELLIPSIS))
  tests.addTests(doctest.DocTestSuite(Output, optionflags=doctest.ELLIPSIS))
  tests.addTests(doctest.DocTestSuite(Supply, optionflags=doctest.ELLIPSIS))
  tests.addTests(doctest.DocTestSuite(Transform, optionflags=doctest.ELLIPSIS))
  return tests


class Test_VLib(unittest.TestCase):

  def test_SupplyNode(self):
    supplier = Supply.SupplyNode()
    self.assertEqual(supplier._contents, [])

    supplier = Supply.SupplyNode(contents=[1, 2, 3])
    self.assertRaises(ValueError, supplier.AddParent, None)

  def test_FilterNode(self):
    filt = Filter.FilterNode(func=lambda a, b: a + b < 5)
    suppl1 = Supply.SupplyNode(contents=[1, 2, 3, 3])
    suppl2 = Supply.SupplyNode(contents=[1, 2, 3, 1])
    filt.AddParent(suppl1)
    filt.AddParent(suppl2)
    self.assertEqual([x for x in filt], [(1, 1), (2, 2), (3, 1)])
    filt.reset()
    self.assertEqual(filt.Negate(), False)
    filt.SetNegate(True)
    self.assertEqual(filt.Negate(), True)
    self.assertEqual([x for x in filt], [
      (3, 3),
    ])
    filt.Destroy()

  def test_OutputNode(self):
    supplier1 = Supply.SupplyNode(contents=[1, 2, 3])
    supplier2 = Supply.SupplyNode(contents=['a', 'b', 'c'])

    sio = StringIO()
    node = Output.OutputNode(dest=sio, strFunc=lambda x: '{0[0]}-{0[1]} '.format(x))
    node.AddParent(supplier1)
    node.AddParent(supplier2)
    result = list(s for s in node)
    self.assertEqual(result, [(1, 'a'), (2, 'b'), (3, 'c')])
    self.assertEqual(sio.getvalue(), '1-a 2-b 3-c ')

    sio = StringIO()
    node = Output.OutputNode(dest=sio)
    node.AddParent(supplier1)
    result = list(s for s in node)
    self.assertEqual(result, [1, 2, 3])
    self.assertEqual(sio.getvalue(), '123')

  def test_VLibNode(self):

    def setupNodes():
      p1 = Node.VLibNode()
      p2 = Node.VLibNode()
      c1 = Node.VLibNode()
      c2 = Node.VLibNode()
      p1.AddChild(c1)
      p2.AddChild(c1)
      p2.AddChild(c2)
      return p1, p2, c1, c2

    p1, p2, c1, c2 = setupNodes()
    # p1 -> c1
    # p2 -> c1
    # p2 -> c2
    self.assertEqual(len(c1.GetParents()), 2)
    self.assertEqual(len(c2.GetParents()), 1)
    self.assertEqual(len(p1.GetChildren()), 1)
    self.assertEqual(len(p2.GetChildren()), 2)

    p1.Destroy()
    # p1
    # p2 -> c1
    # p2 -> c2
    self.assertEqual(len(c1.GetParents()), 1)
    self.assertEqual(len(c2.GetParents()), 1)
    self.assertEqual(len(p1.GetChildren()), 0)
    self.assertEqual(len(p2.GetChildren()), 2)

    p1, p2, c1, c2 = setupNodes()
    p1.Destroy(propagateDown=True)
    # p1, c1
    # p2 -> c2
    self.assertEqual(len(c1.GetParents()), 0)
    self.assertEqual(len(c2.GetParents()), 1)
    self.assertEqual(len(p1.GetChildren()), 0)
    self.assertEqual(len(p2.GetChildren()), 1)

    p1, p2, c1, c2 = setupNodes()
    p1.Destroy(propagateUp=True)
    # p1
    # p2 -> c1
    # p2 -> c2
    self.assertEqual(len(c1.GetParents()), 1)
    self.assertEqual(len(c2.GetParents()), 1)
    self.assertEqual(len(p1.GetChildren()), 0)
    self.assertEqual(len(p2.GetChildren()), 2)

    p1, p2, c1, c2 = setupNodes()
    c1.Destroy(propagateUp=True)
    # p1, c1, p2, c2
    self.assertEqual(len(c1.GetParents()), 0)
    self.assertEqual(len(c2.GetParents()), 0)
    self.assertEqual(len(p1.GetChildren()), 0)
    self.assertEqual(len(p2.GetChildren()), 0)

    p1, p2, c1, c2 = setupNodes()
    p1.Destroy(propagateDown=True)
    # p1, c1
    # p2 -> c2
    self.assertEqual(len(c1.GetParents()), 0)
    self.assertEqual(len(c2.GetParents()), 1)
    self.assertEqual(len(p1.GetChildren()), 0)
    self.assertEqual(len(p2.GetChildren()), 1)
    p1.Destroy(propagateDown=True)
    # p1, c1
    # p2 -> c2
    self.assertEqual(len(c1.GetParents()), 0)
    self.assertEqual(len(c2.GetParents()), 1)
    self.assertEqual(len(p1.GetChildren()), 0)
    self.assertEqual(len(p2.GetChildren()), 1)


if __name__ == '__main__':  # pragma: nocover
  unittest.main()



#! make_templates.py
"""Parse a C++ input and try to generate SWIG %template declarations so that this
    template <typename T1, typename T2>
    DoubleVect
    OnBitProjSimilarity(const T1& bv1,const T2& bv2);
becomes this
    %template(OnBitProjSimilarityEBV) OnBitProjSimilarity<ExplicitBitVect,ExplicitBitVect>;
"""

import re

template_match = re.compile(r"""template\s*\<(.+)\>\s*.*\s+(\w+)\s*\(.*""")


def make_entry(name, suffix, template_types):
  template_part = '<' + ','.join(template_types) + '>'
  return '%template(' + name + suffix + ') ' + name + template_part + ';'


def handle_template_line(template_line, suffix, template_type):
  # print template_line
  match = template_match.search(template_line)
  types = match.group(1)
  name = match.group(2)
  template_types = len(types.split(',')) * [template_type]
  print(make_entry(name, suffix, template_types))


def find_declarations(file_name, suffix, template_type):
  f = open(file_name)
  in_template = False
  template_line = ''
  for line in f:
    if in_template:
      template_line += ' ' + line.strip()
      if line.find(')') >= 0:
        handle_template_line(template_line, suffix, template_type)
        in_template = False
    elif line.strip().startswith('template'):
      in_template = True
      template_line = line.strip()
  f.close()



# $Id$
#
import unittest

import numpy

from rdkit import Chem, RDConfig
from rdkit.Chem import rdReducedGraphs as rdRG


class TestCase(unittest.TestCase):

  def setUp(self):
    pass

  def test1(self):
    m = Chem.MolFromSmiles('OCCc1ccccc1')
    mrg = rdRG.GenerateMolExtendedReducedGraph(m)
    mrg.UpdatePropertyCache(False)
    self.assertEqual('*cCCO', Chem.MolToSmiles(mrg))

    m = Chem.MolFromSmiles('OCCC1CCCCC1')
    mrg = rdRG.GenerateMolExtendedReducedGraph(m)
    mrg.UpdatePropertyCache(False)
    self.assertEqual('*CCCO', Chem.MolToSmiles(mrg))

  def test2(self):
    m = Chem.MolFromSmiles('OCCc1ccccc1')
    mrg = rdRG.GenerateMolExtendedReducedGraph(m)
    mrg.UpdatePropertyCache(False)
    self.assertEqual('*cCCO', Chem.MolToSmiles(mrg))

    fp1 = rdRG.GenerateErGFingerprintForReducedGraph(mrg)
    fp2 = rdRG.GetErGFingerprint(m)
    md = max(abs(fp1 - fp2))
    self.assertLess(md, 1e-4)

  def test3(self):
    m = Chem.MolFromSmiles('OCCc1ccccc1')
    fp1 = rdRG.GetErGFingerprint(m)
    m = Chem.MolFromSmiles('OCCC1CC=CC=C1')
    fp2 = rdRG.GetErGFingerprint(m)

    md = max(abs(fp1 - fp2))
    self.assertAlmostEqual(0.0, md, 4)

  def test4(self):
    m = Chem.MolFromSmiles('OCCc1ccccc1')
    fp1 = rdRG.GetErGFingerprint(m)
    fp2 = rdRG.GetErGFingerprint(m, fuzzIncrement=0.1)

    md = max(abs(fp1 - fp2))
    self.assertAlmostEqual(0.2, md, 4)

  def testCanRetrieveProp(self):
    m = Chem.MolFromSmiles('OCCc1ccccc1')
    mrg = rdRG.GenerateMolExtendedReducedGraph(m)
    erg_types = [tuple(atom.GetPropsAsDict().get('_ErGAtomTypes')) for atom in mrg.GetAtoms()]
    self.assertEqual(erg_types, [(0, 1), (), (), (), (5, )])


if __name__ == '__main__':
  unittest.main()



#
#  Copyright (C) 2004-2008  greg Landrum and Rational Discovery LLC
#
#   @@ All Rights Reserved @@
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#

import doctest
import gzip
import io
import os
import pickle
import unittest

from rdkit import Chem
from rdkit import DistanceGeometry as DG
from rdkit import Geometry, RDConfig
from rdkit.Chem import ChemicalFeatures, rdDistGeom
from rdkit.Chem.Pharm3D import EmbedLib, Pharmacophore


def feq(n1, n2, tol=1e-5):
  return abs(n1 - n2) <= tol


def load_tests(loader, tests, ignore):
  """ Add the Doctests from the module """
  tests.addTests(
    doctest.DocTestSuite(EmbedLib, optionflags=doctest.ELLIPSIS + doctest.NORMALIZE_WHITESPACE))
  return tests


class TestCase(unittest.TestCase):

  def setUp(self):
    self.dataDir = os.path.join(RDConfig.RDCodeDir, 'Chem/Pharm3D/test_data')
    self.fdefBlock = """
                   DefineFeature HAcceptor1 [N,O;H0]
                      Family HBondAcceptor
                      Weights 1.0
                   EndFeature
                   DefineFeature HDonor1 [N,O;!H0]
                      Family HBondDonor
                      Weights 1.0
                   EndFeature
                   DefineFeature Aromatic1 c1ccccc1
                      Family Aromatic
                      Weights 1.,1.,1.,1.,1.,1.
                   EndFeature\n"""

    self.featFactory = ChemicalFeatures.BuildFeatureFactoryFromString(self.fdefBlock)
    self.feats = [
      ChemicalFeatures.FreeChemicalFeature('HBondAcceptor', 'HAcceptor1',
                                           Geometry.Point3D(0.0, 0.0, 0.0)),
      ChemicalFeatures.FreeChemicalFeature('HBondDonor', 'HDonor1',
                                           Geometry.Point3D(2.65, 0.0, 0.0)),
      ChemicalFeatures.FreeChemicalFeature('Aromatic', 'Aromatic1',
                                           Geometry.Point3D(5.12, 0.908, 0.0)),
    ]
    self.pcophore = Pharmacophore.Pharmacophore(self.feats)
    self.pcophore.setLowerBound(0, 1, 2.0)
    self.pcophore.setUpperBound(0, 1, 3.3)
    self.pcophore.setLowerBound(0, 2, 5.0)
    self.pcophore.setUpperBound(0, 2, 5.4)
    self.pcophore.setLowerBound(1, 2, 2.6)
    self.pcophore.setUpperBound(1, 2, 3.0)

  def _matchMol(self, tpl, pcophore, featFactory, downSample):
    _, molPkl, boundsMat = tpl
    mol = Chem.Mol(molPkl)
    matched, matches = EmbedLib.MatchPharmacophoreToMol(mol, featFactory, pcophore)
    if matched:
      r = EmbedLib.MatchPharmacophore(matches, boundsMat, pcophore, useDownsampling=downSample)
      return int(not bool(r[0]))
    else:
      return 0

  def test1SearchFullMat(self):
    inF = gzip.open(os.path.join(self.dataDir, 'cdk2-syn-clip100.pkl.gz'), 'rb')
    # outF = gzip.open(os.path.join(self.dataDir,'cdk2-syn-clip100.pkl.new.gz'),'wb+')
    nDone = 0
    nHits = 0
    while 1:
      try:
        tpl = pickle.load(inF, encoding='latin1')
        tpl = tpl[0], tpl[1].encode('latin1'), tpl[2]
        # tpl=tpl[0],tpl[1],numpy.array(tpl[2])
        # pickle.dump(tpl,outF)
      except Exception:
        break
      if self._matchMol(tpl, self.pcophore, self.featFactory, 0):
        nHits += 1
      nDone += 1
    self.assertEqual(nDone, 100)
    # print 'nHits:',nHits
    self.assertEqual(nHits, 47)

  def test2SearchDownsample(self):
    inF = gzip.open(os.path.join(self.dataDir, 'cdk2-syn-clip100.pkl.gz'), 'rb')
    nDone = 0
    nHits = 0
    while 1:
      try:
        tpl = pickle.load(inF, encoding='latin1')
        tpl = tpl[0], tpl[1].encode('latin1'), tpl[2]
      except Exception:
        break
      if self._matchMol(tpl, self.pcophore, self.featFactory, 1):
        nHits += 1
      nDone += 1
    self.assertEqual(nDone, 100)
    # print 'nHits:',nHits
    self.assertEqual(nHits, 47)

  def test3Embed(self):
    testResults = {
      'mol_197': (219.10, 35.89, 110.98, 11.56, 110.32, 11.04, 95.23, 10.25, 0.00),
      'mol_223': (243.67, 7.24, 134.74, 1.55, 134.66, 1.55, 85.27, 0.61, 0.00),
      'mol_269': (215.52, 6.07, 103.94, 0.21, 102.76, 0.27, 88.22, 0.36, 6.00),
    }
    inF = gzip.open(os.path.join(self.dataDir, 'cdk2-syn-clip100.pkl.gz'), 'rb')
    nDone = 0
    nHits = 0
    while 1:
      try:
        name, molPkl, _ = pickle.load(inF, encoding='latin1')
        molPkl = bytes(molPkl, encoding='latin1')
      except Exception:
        break

      nDone += 1

      mol = Chem.Mol(molPkl)
      nboundsMat = rdDistGeom.GetMoleculeBoundsMatrix(mol)
      DG.DoTriangleSmoothing(nboundsMat)
      matched, matches = EmbedLib.MatchPharmacophoreToMol(mol, self.featFactory, self.pcophore)
      if matched:
        failed, _, match, stats = EmbedLib.MatchPharmacophore(matches, nboundsMat, self.pcophore,
                                                              useDownsampling=1)
        if not failed:
          nHits += 1

          if name in testResults:
            stats = EmbedLib.EmbedOne(mol, name, match, self.pcophore, count=10, silent=1,
                                      randomSeed=23)
            tgt = testResults[name]
            self.assertEqual(len(tgt), len(stats))
            print(name)
            print(','.join([f'{x:.2f}' for x in stats]))
            # we'll use different tolerances for the different values:
            self.assertTrue(feq(tgt[0], stats[0], 5.0), (tgt[0], stats[0]))
            for i in range(2, len(tgt)):
              self.assertTrue(feq(tgt[i], stats[i], 5.0), (tgt[i], stats[i]))

    self.assertEqual(nDone, 100)
    # print 'nHits:',nHits
    self.assertEqual(nHits, 50)

  def test4Search(self):
    featFactory = ChemicalFeatures.BuildFeatureFactory(
      os.path.join(self.dataDir, 'BaseFeatures.fdef'))

    activeFeats = [
      ChemicalFeatures.FreeChemicalFeature('Acceptor', Geometry.Point3D(0.0, 0.0, 0.0)),
      ChemicalFeatures.FreeChemicalFeature('Donor', Geometry.Point3D(0.0, 0.0, 0.0)),
      ChemicalFeatures.FreeChemicalFeature('Aromatic', Geometry.Point3D(0.0, 0.0, 0.0))
    ]
    pcophore = Pharmacophore.Pharmacophore(activeFeats)
    pcophore.setLowerBound(0, 1, 2.251)
    pcophore.setUpperBound(0, 1, 2.451)
    pcophore.setUpperBound2D(0, 1, 3)

    pcophore.setLowerBound(0, 2, 4.970)
    pcophore.setUpperBound(0, 2, 5.170)
    pcophore.setUpperBound2D(0, 2, 6)

    pcophore.setLowerBound(1, 2, 2.681)
    pcophore.setUpperBound(1, 2, 2.881)
    pcophore.setUpperBound2D(1, 2, 6)

    inF = gzip.open(os.path.join(self.dataDir, 'cdk2-syn-clip100.pkl.gz'), 'rb')
    nDone = 0
    nMatches = 0
    nHits = 0

    while 1:
      try:
        _, molPkl, boundsMat = pickle.load(inF, encoding='latin1')
        molPkl = bytes(molPkl, encoding='latin1')
      except Exception:
        break

      nDone += 1

      mol = Chem.Mol(molPkl)
      boundsMat = rdDistGeom.GetMoleculeBoundsMatrix(mol)
      DG.DoTriangleSmoothing(boundsMat)

      canMatch, matches = EmbedLib.MatchPharmacophoreToMol(mol, featFactory, pcophore)
      if canMatch:
        nMatches += 1
        r = EmbedLib.MatchPharmacophore(matches, boundsMat, pcophore, useDownsampling=True,
                                        use2DLimits=True, mol=mol)
        failed = r[0]
        if not failed:
          nHits += 1

    self.assertEqual(nDone, 100)
    self.assertEqual(nMatches, 93)
    # print 'nhits:',nHits
    self.assertEqual(nHits, 67)

  def testIssue268(self):
    featFactory = ChemicalFeatures.BuildFeatureFactory(os.path.join(self.dataDir, 'Issue268.fdef'))
    m1 = Chem.MolFromMolFile(os.path.join(self.dataDir, 'Issue268_Mol1.mol'))
    m2 = Chem.MolFromMolFile(os.path.join(self.dataDir, 'Issue268_Mol2.mol'))
    with open(os.path.join(self.dataDir, 'Issue268_Pcop.pkl'), 'r') as inTF:
      buf = inTF.read().replace('\r\n', '\n').encode('utf-8')
      inTF.close()
    with io.BytesIO(buf) as inF:
      pcop = pickle.load(inF, encoding='latin1')
    # pcop._boundsMat=numpy.array(pcop._boundsMat)
    # pcop._boundsMat2D=numpy.array(pcop._boundsMat2D)
    # pickle.dump(pcop,file(os.path.join(self.dataDir,
    #                                    'Issue268_Pcop.new.pkl'),'wb+'))
    _, mList1 = EmbedLib.MatchFeatsToMol(m1, featFactory, pcop.getFeatures())
    _, mList2 = EmbedLib.MatchFeatsToMol(m2, featFactory, pcop.getFeatures())
    b1 = rdDistGeom.GetMoleculeBoundsMatrix(m1)
    b2 = rdDistGeom.GetMoleculeBoundsMatrix(m2)

    self.assertEqual(len(EmbedLib.MatchPharmacophore(mList1, b1, pcop)[2]), 4)
    self.assertEqual(len(EmbedLib.MatchPharmacophore(mList2, b2, pcop)[2]), 4)

    self.assertEqual(
      len(EmbedLib.MatchPharmacophore(mList1, b1, pcop, mol=m1, use2DLimits=True)[2]), 4)
    self.assertEqual(
      len(EmbedLib.MatchPharmacophore(mList2, b2, pcop, mol=m2, use2DLimits=True)[2]), 4)

    self.assertTrue(DG.DoTriangleSmoothing(b1))
    self.assertTrue(DG.DoTriangleSmoothing(b2))

    self.assertEqual(len(EmbedLib.MatchPharmacophore(mList1, b1, pcop)[2]), 4)
    self.assertEqual(len(EmbedLib.MatchPharmacophore(mList2, b2, pcop)[2]), 4)

    self.assertEqual(
      len(EmbedLib.MatchPharmacophore(mList1, b1, pcop, mol=m1, use2DLimits=True)[2]), 4)
    self.assertEqual(
      len(EmbedLib.MatchPharmacophore(mList2, b2, pcop, mol=m2, use2DLimits=True)[2]), 4)


if __name__ == '__main__':  # pragma: nocover
  unittest.main()



import unittest

from rdkit import Chem, RDConfig
from rdkit.Chem import rdMolInterchange


class TestCase(unittest.TestCase):

  def setUp(self):
    pass

  def test1(self):
    smis = ('c1ccccc1', 'C[C@H](F)Cl')
    for smi in smis:
      m = Chem.MolFromSmiles(smi)
      csmi = Chem.MolToSmiles(m)
      json = rdMolInterchange.MolToJSON(m)
      nms = rdMolInterchange.JSONToMols(json)
      self.assertEqual(len(nms), 1)
      smi2 = Chem.MolToSmiles(nms[0])
      self.assertEqual(csmi, smi2)
    ms = [Chem.MolFromSmiles(smi) for smi in smis]
    json = rdMolInterchange.MolsToJSON(ms)
    nms = rdMolInterchange.JSONToMols(json)
    self.assertEqual(len(ms), len(nms))
    self.assertEqual([Chem.MolToSmiles(x) for x in ms], [Chem.MolToSmiles(x) for x in nms])

  def test2(self):
    smis = ("C[C@H](O)C[C@@H](C)F |o1:1,4|", "C[C@H](O)CC[C@@H](C)F |&1:1,5|")
    ms = [Chem.MolFromSmiles(x) for x in smis]
    json = rdMolInterchange.MolToJSON(ms[0])
    self.assertIn('stereoGroups', json)
    self.assertIn('"or"', json)
    self.assertIn('[1,4]', json)

    json = rdMolInterchange.MolToJSON(ms[1])
    self.assertIn('stereoGroups', json)
    self.assertIn('"and"', json)
    self.assertIn('[1,5]', json)

    json = rdMolInterchange.MolsToJSON(ms)
    self.assertIn('stereoGroups', json)
    self.assertIn('"or"', json)
    self.assertIn('[1,4]', json)
    self.assertIn('"and"', json)
    self.assertIn('[1,5]', json)

    ps = rdMolInterchange.JSONWriteParameters()
    ps.useRDKitExtensions = False
    json = rdMolInterchange.MolToJSON(ms[1], ps)
    self.assertNotIn('stereoGroups', json)
    json = rdMolInterchange.MolsToJSON(ms, ps)
    self.assertNotIn('stereoGroups', json)


if __name__ == '__main__':
  unittest.main()



# $Id$
#
#  Copyright (C) 2001-2008  greg Landrum and Rational Discovery LLC
#
#   @@ All Rights Reserved @@
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#
"""unit testing code for clustering

"""

import unittest
from io import StringIO

import numpy

from rdkit.ML.Cluster import Clusters, ClusterUtils, Murtagh
from rdkit.TestRunner import redirect_stdout


class TestCase(unittest.TestCase):

  def setUp(self):
    # this is the data set used by Romesburg in "Cluster Analysis for Researchers"
    #  to demonstrate the different clustering methods
    # print '\n%s: '%self.shortDescription(),
    self.d = numpy.array([[10., 5.], [20., 20.], [30., 10.], [30., 15.], [5., 10.]])
    self.names = ['p1', 'p2', 'p3', 'p4', 'p5']

  def testDivide(self):
    " tests the cluster division algorithms "
    ca = Clusters.Cluster(index=1)
    cb = Clusters.Cluster(index=2)
    cc = Clusters.Cluster(index=3)
    cd = Clusters.Cluster(index=4)
    ce = Clusters.Cluster(index=5)
    cf = Clusters.Cluster(index=6)

    c1 = Clusters.Cluster(metric=10, children=[ca, cb], index=7)
    c2 = Clusters.Cluster(metric=15, children=[cc, cd], index=8)
    c3 = Clusters.Cluster(metric=20, children=[ce, cf], index=9)
    c4 = Clusters.Cluster(metric=25, children=[c2, c3], index=10)
    c5 = Clusters.Cluster(metric=30, children=[c4, c1], index=11)

    cs = ClusterUtils.SplitIntoNClusters(c5, 4, breadthFirst=True)
    assert len(cs) == 4, 'bad split length'
    indices = [x.GetIndex() for x in cs]
    for index in [9, 8, 1, 2]:
      assert index in indices, 'index %d not found in %s' % (index, str(indices))
    # we may not want to preserve order, but test it for now
    assert indices == [9, 8, 1, 2], 'bad index order'

    cs2 = ClusterUtils.SplitIntoNClusters(c5, 4, breadthFirst=False)
    indices = [x.GetIndex() for x in cs2]
    for index in [8, 7, 5, 6]:
      assert index in indices, 'index %d not found in %s' % (index, str(indices))
    # we may not want to preserve order, but test it for now
    assert indices == [8, 7, 5, 6], 'bad index order'

    # Exceptions and edge cases
    self.assertRaises(ValueError, ClusterUtils.SplitIntoNClusters, c5, len(c5) + 1)
    self.assertEqual(ClusterUtils.SplitIntoNClusters(c5, len(c5)), c5.GetPoints())
    self.assertEqual(ClusterUtils.SplitIntoNClusters(c5, 0), [c5])

    for n in range(len(c5)):
      if n >= 7:  # Code fails for n = 7 and above
        self.assertRaises(AssertionError, ClusterUtils.SplitIntoNClusters, c5, n, breadthFirst=True)
      else:
        ClusterUtils.SplitIntoNClusters(c5, n, breadthFirst=True)

    self.assertRaises(ValueError, ClusterUtils.SplitIntoNClusters, c5,
                      len(c5) + 1, breadthFirst=False)
    self.assertEqual(ClusterUtils.SplitIntoNClusters(c5, len(c5), breadthFirst=False),
                     c5.GetPoints())
    self.assertEqual(ClusterUtils.SplitIntoNClusters(c5, 0, breadthFirst=False), [c5])

    for n in range(len(c5)):
      if n >= 7:  # Code fails for n = 7 and above
        self.assertRaises(AssertionError, ClusterUtils.SplitIntoNClusters, c5, n,
                          breadthFirst=False)
      else:
        ClusterUtils.SplitIntoNClusters(c5, n, breadthFirst=False)

  @unittest.skipIf(Murtagh.MurtaghCluster is None, "Murtagh clustering not available")
  def testMurtaghUPGMA(self):
    if Murtagh is None:
      return
    nPts = 5
    sz = 5
    dataP = numpy.random.random((nPts, sz))
    newClust = Murtagh.ClusterData(dataP, nPts, Murtagh.UPGMA)[0]
    ds = []
    for i in range(nPts):
      for j in range(i):
        d = dataP[i] - dataP[j]
        ds.append(sum(d * d))
    ds = numpy.array(ds)
    newClust2 = Murtagh.ClusterData(ds, nPts, Murtagh.UPGMA, isDistData=1)[0]

    assert len(newClust) == len(newClust2), 'length mismatch2'

    assert not newClust.Compare(newClust2, ignoreExtras=0), 'equality failed3'

    newClust2 = Murtagh.ClusterData(dataP, nPts, Murtagh.UPGMA, isDistData=0)[0]
    assert len(newClust) == len(newClust2), 'length mismatch2'

    assert not newClust.Compare(newClust2, ignoreExtras=0), 'equality failed3'

  def test_Cluster(self):
    """ tests the Cluster class functionality """
    root = Clusters.Cluster(index=1, position=1)
    c1 = Clusters.Cluster(index=10, position=10)
    c1.AddChild(Clusters.Cluster(index=30, position=30))
    c1.AddChild(Clusters.Cluster(index=31, position=31))
    t32 = Clusters.Cluster(index=32, position=32)
    c1.AddChild(t32)

    c2 = Clusters.Cluster(index=11)
    #     c2.AddChild(Clusters.Cluster(index=40))
    #     c2.AddChild(Clusters.Cluster(index=41))
    c2.AddChildren([Clusters.Cluster(index=40), Clusters.Cluster(index=41)])

    root.AddChild(c1)
    root.AddChild(c2)
    nodes = ClusterUtils.GetNodeList(root)

    indices = [x.GetIndex() for x in nodes]
    assert indices == [30, 31, 32, 10, 40, 41, 11, 1], 'bad indices'
    subtree = root.FindSubtree(11)
    self.assertEqual([x.GetIndex() for x in ClusterUtils.GetNodeList(subtree)], [40, 41, 11])

    self.assertFalse(root.IsTerminal())
    self.assertTrue(t32.IsTerminal())

    self.assertEqual(root.GetData(), None)
    root.SetData(3.14)
    self.assertEqual(root.GetData(), 3.14)

    self.assertEqual(root.GetMetric(), 0.0)
    root.SetMetric(0.1)
    self.assertEqual(root.GetMetric(), 0.1)

    self.assertEqual(root.GetIndex(), 1)
    root.SetIndex(100)
    self.assertEqual(root.GetIndex(), 100)

    self.assertEqual(root.GetPointsPositions(), [30, 31, 32, []])

    root.RemoveChild(c1)
    self.assertEqual([x.GetIndex() for x in ClusterUtils.GetNodeList(root)], [40, 41, 11, 100])

    self.assertEqual(root.GetName(), 'Cluster(100)')
    root.SetName('abc')
    self.assertEqual(root.GetName(), 'abc')

    f = StringIO()
    with redirect_stdout(f):
      root.Print(showData=True)
    self.assertIn('abc', f.getvalue())
    self.assertIn('Cluster(41)', f.getvalue())
    self.assertIn('Metric', f.getvalue())


if __name__ == '__main__':  # pragma: nocover
  unittest.main()



import python_objs


def lentest():
  assert python_objs.seq_len([1, 2]) == 2
  assert python_objs.seq_len((1, 2)) == 2
  assert python_objs.seq_len("fooo") == 4
  try:
    python_objs.seq_len(1)
  except TypeError:
    pass
  except Exception:
    assert 0, 'wrong exception type'
  else:
    assert 0, 'no exception'


def sumtest():
  assert python_objs.sum_first2([1, 2]) == 3
  assert python_objs.sum_first2((1, 3)) == 4
  try:
    python_objs.sum_first2((1., 3.))
  except TypeError:
    pass
  except Exception:
    assert 0, 'wrong exception type'
  else:
    assert 0, 'no exception'
  try:
    python_objs.sum_first2('foo')
  except TypeError:
    pass
  except Exception:
    assert 0, 'wrong exception type'
  else:
    assert 0, 'no exception'
  try:
    python_objs.sum_first2(1)
  except TypeError:
    pass
  except Exception:
    assert 0, 'wrong exception type'
  else:
    assert 0, 'no exception'


lentest()
sumtest()



# $Id$
#
#  Copyright (C) 2007  greg Landrum
#
#   @@ All Rights Reserved @@
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#

import doctest
import unittest

from rdkit import RDConfig
from rdkit.Chem.Fingerprints import DbFpSupplier
from rdkit.Dbase.DbConnection import DbConnect


def load_tests(loader, tests, ignore):
  """ Add the Doctests from the module """
  tests.addTests(doctest.DocTestSuite(DbFpSupplier, optionflags=doctest.ELLIPSIS))
  return tests


class TestCase(unittest.TestCase):

  def test_ForwardDbFpSupplier(self):
    # Additional tests to complete code coverage
    conn = DbConnect(RDConfig.RDTestDatabase, 'simple_combined')

    self.assertRaises(ValueError, DbFpSupplier.ForwardDbFpSupplier, conn.GetData(),
                      fpColName='typo')

    suppl = DbFpSupplier.ForwardDbFpSupplier(conn.GetData(), fpColName='AutoFragmentFp')
    self.assertIn('ID', suppl.GetColumnNames())



# $Id$
#
# Copyright (C) 2007,2008 Greg Landrum
#
#  @@ All Rights Reserved @@
#
import io
import os
import pickle
import unittest

from rdkit import DataStructs as ds
from rdkit import RDConfig


class TestCase(unittest.TestCase):

  def setUp(self):
    pass

  def test1Discrete(self):
    v1 = ds.DiscreteValueVect(ds.DiscreteValueType.ONEBITVALUE, 30)
    for i in range(15):
      v1[2 * i] = 1

    self.assertTrue(len(v1) == 30)
    self.assertTrue(v1.GetTotalVal() == 15)

    for i, vv1 in enumerate(v1):
      self.assertTrue(vv1 == (i + 1) % 2)

    self.assertRaises(ValueError, lambda: v1.__setitem__(5, 2))

    v1 = ds.DiscreteValueVect(ds.DiscreteValueType.TWOBITVALUE, 30)
    for i, _ in enumerate(v1):
      v1[i] = i % 4

    self.assertTrue(len(v1) == 30)
    for i, vv1 in enumerate(v1):
      self.assertTrue(vv1 == i % 4)

    self.assertRaises(ValueError, lambda: v1.__setitem__(10, 6))

    v1 = ds.DiscreteValueVect(ds.DiscreteValueType.FOURBITVALUE, 30)
    for i, _ in enumerate(v1):
      v1[i] = i % 16

    self.assertTrue(len(v1) == 30)
    self.assertTrue(v1.GetTotalVal() == 211)
    for i, vv1 in enumerate(v1):
      self.assertTrue(vv1 == i % 16)

    self.assertRaises(ValueError, lambda: v1.__setitem__(10, 16))

    v1 = ds.DiscreteValueVect(ds.DiscreteValueType.EIGHTBITVALUE, 32)
    for i, _ in enumerate(v1):
      v1[i] = i % 256

    self.assertTrue(len(v1) == 32)
    self.assertTrue(v1.GetTotalVal() == 496)
    for i, vv1 in enumerate(v1):
      self.assertTrue(vv1 == i % 256)

    self.assertRaises(ValueError, lambda: v1.__setitem__(10, 256))

    v1 = ds.DiscreteValueVect(ds.DiscreteValueType.SIXTEENBITVALUE, 300)
    for i, _ in enumerate(v1):
      v1[i] = i % 300

    self.assertTrue(len(v1) == 300)
    self.assertTrue(v1.GetTotalVal() == 44850)
    self.assertRaises(ValueError, lambda: v1.__setitem__(10, 65536))

  def test2VectDistances(self):
    v1 = ds.DiscreteValueVect(ds.DiscreteValueType.ONEBITVALUE, 30)
    v2 = ds.DiscreteValueVect(ds.DiscreteValueType.ONEBITVALUE, 30)
    for i in range(15):
      v1[2 * i] = 1
      v2[2 * i] = 1
    self.assertTrue(ds.ComputeL1Norm(v1, v2) == 0)
    for i in range(30):
      if (i % 3 == 0):
        v2[i] = 1
      else:
        v2[i] = 0
    self.assertTrue(ds.ComputeL1Norm(v1, v2) == 15)

    v1 = ds.DiscreteValueVect(ds.DiscreteValueType.TWOBITVALUE, 30)
    v2 = ds.DiscreteValueVect(ds.DiscreteValueType.TWOBITVALUE, 30)

    for i in range(30):
      v1[i] = i % 4
      v2[i] = (i + 1) % 4

    self.assertTrue(ds.ComputeL1Norm(v1, v2) == 44)

    v1 = ds.DiscreteValueVect(ds.DiscreteValueType.FOURBITVALUE, 16)
    v2 = ds.DiscreteValueVect(ds.DiscreteValueType.FOURBITVALUE, 16)
    for i in range(16):
      v1[i] = i % 16
      v2[i] = i % 5
    self.assertTrue(ds.ComputeL1Norm(v1, v2) == 90)

    v1 = ds.DiscreteValueVect(ds.DiscreteValueType.EIGHTBITVALUE, 5)
    v2 = ds.DiscreteValueVect(ds.DiscreteValueType.EIGHTBITVALUE, 5)
    v1[0] = 34
    v1[1] = 167
    v1[2] = 3
    v1[3] = 56
    v1[4] = 128

    v2[0] = 14
    v2[1] = 67
    v2[2] = 103
    v2[3] = 6
    v2[4] = 228
    self.assertTrue(ds.ComputeL1Norm(v1, v2) == 370)

    v1 = ds.DiscreteValueVect(ds.DiscreteValueType.SIXTEENBITVALUE, 3)
    v2 = ds.DiscreteValueVect(ds.DiscreteValueType.SIXTEENBITVALUE, 3)
    v1[0] = 2345
    v1[1] = 64578
    v1[2] = 34

    v2[0] = 1345
    v2[1] = 54578
    v2[2] = 10034
    self.assertTrue(ds.ComputeL1Norm(v1, v2) == 21000)

  def test3Pickles(self):
    #outF = file('dvvs.pkl','wb+')
    with open(os.path.join(RDConfig.RDBaseDir, 'Code/DataStructs/Wrap/testData/dvvs.pkl'),
              'r') as inTF:
      buf = inTF.read().replace('\r\n', '\n').encode('utf-8')
      inTF.close()
    with io.BytesIO(buf) as inF:
      v1 = ds.DiscreteValueVect(ds.DiscreteValueType.ONEBITVALUE, 30)
      for i in range(15):
        v1[2 * i] = 1
      v2 = pickle.loads(pickle.dumps(v1))
      self.assertTrue(ds.ComputeL1Norm(v1, v2) == 0)
      # pickle.dump(v1,outF)
      v2 = pickle.load(inF, encoding='bytes')
      self.assertTrue(ds.ComputeL1Norm(v1, v2) == 0)
      self.assertTrue(v1.GetTotalVal() == v2.GetTotalVal())
      self.assertTrue(v2.GetTotalVal() != 0)

      v1 = ds.DiscreteValueVect(ds.DiscreteValueType.TWOBITVALUE, 30)
      for i in range(30):
        v1[i] = i % 4
      v2 = pickle.loads(pickle.dumps(v1))
      self.assertTrue(ds.ComputeL1Norm(v1, v2) == 0)
      # pickle.dump(v1,outF)
      v2 = pickle.load(inF, encoding='bytes')
      self.assertTrue(ds.ComputeL1Norm(v1, v2) == 0)
      self.assertTrue(v1.GetTotalVal() == v2.GetTotalVal())
      self.assertTrue(v2.GetTotalVal() != 0)

      v1 = ds.DiscreteValueVect(ds.DiscreteValueType.FOURBITVALUE, 16)
      for i in range(16):
        v1[i] = i % 16
      v2 = pickle.loads(pickle.dumps(v1))
      self.assertTrue(ds.ComputeL1Norm(v1, v2) == 0)
      # pickle.dump(v1,outF)
      v2 = pickle.load(inF, encoding='bytes')
      self.assertTrue(ds.ComputeL1Norm(v1, v2) == 0)
      self.assertTrue(v1.GetTotalVal() == v2.GetTotalVal())
      self.assertTrue(v2.GetTotalVal() != 0)

      v1 = ds.DiscreteValueVect(ds.DiscreteValueType.EIGHTBITVALUE, 5)
      v1[0] = 34
      v1[1] = 167
      v1[2] = 3
      v1[3] = 56
      v1[4] = 128
      v2 = pickle.loads(pickle.dumps(v1))
      self.assertTrue(ds.ComputeL1Norm(v1, v2) == 0)
      # pickle.dump(v1,outF)
      v2 = pickle.load(inF, encoding='bytes')
      self.assertTrue(ds.ComputeL1Norm(v1, v2) == 0)
      self.assertTrue(v1.GetTotalVal() == v2.GetTotalVal())
      self.assertTrue(v2.GetTotalVal() != 0)

      v1 = ds.DiscreteValueVect(ds.DiscreteValueType.SIXTEENBITVALUE, 3)
      v1[0] = 2345
      v1[1] = 64578
      v1[2] = 34
      v2 = pickle.loads(pickle.dumps(v1))
      self.assertTrue(ds.ComputeL1Norm(v1, v2) == 0)
      # pickle.dump(v1,outF)
      v2 = pickle.load(inF, encoding='bytes')
      self.assertTrue(ds.ComputeL1Norm(v1, v2) == 0)
      self.assertTrue(v1.GetTotalVal() == v2.GetTotalVal())
      self.assertTrue(v2.GetTotalVal() != 0)

  def test4DiscreteVectOps(self):
    v1 = ds.DiscreteValueVect(ds.DiscreteValueType.TWOBITVALUE, 8)
    for i in range(4):
      v1[2 * i] = 2
    self.assertTrue(v1.GetTotalVal() == 8)
    v2 = ds.DiscreteValueVect(ds.DiscreteValueType.TWOBITVALUE, 8)
    for i in range(4):
      v2[2 * i + 1] = 2
      v2[2 * i] = 1
    self.assertTrue(v2.GetTotalVal() == 12)

    v3 = v1 | v2
    self.assertTrue(len(v3) == len(v2))
    self.assertTrue(v3.GetTotalVal() == 16)

    v3 = v1 & v2
    self.assertTrue(len(v3) == len(v2))
    self.assertTrue(v3.GetTotalVal() == 4)

    v4 = v1 + v2
    self.assertTrue(len(v4) == len(v2))
    self.assertTrue(v4.GetTotalVal() == 20)

    v4 = v1 - v2
    self.assertTrue(v4.GetTotalVal() == 4)
    v4 = v2 - v1
    self.assertTrue(v4.GetTotalVal() == 8)

    v4 = v2
    v4 -= v1
    self.assertTrue(v4.GetTotalVal() == 8)
    v4 -= v4
    self.assertTrue(v4.GetTotalVal() == 0)

  def testIterator(self):
    """
        connected to sf.net issue 1719831:
        http://sourceforge.net/tracker/index.php?func=detail&aid=1719831&group_id=160139&atid=814650
        """
    v1 = ds.DiscreteValueVect(ds.DiscreteValueType.ONEBITVALUE, 30)
    for i in range(15):
      v1[2 * i] = 1
    l1 = list(v1)
    self.assertTrue(len(l1) == len(v1))
    for v, l in zip(v1, l1):
      self.assertTrue(l == v)
    self.assertRaises(IndexError, lambda: v1[40])

  def test9ToNumpy(self):
    import numpy
    bv = ds.DiscreteValueVect(ds.DiscreteValueType.FOURBITVALUE, 32)
    bv[0] = 1
    bv[1] = 4
    bv[17] = 1
    bv[23] = 8
    bv[31] = 12
    arr = numpy.zeros((3, ), 'i')
    ds.ConvertToNumpyArray(bv, arr)
    self.assertTrue(numpy.all(bv == arr))


if __name__ == '__main__':
  unittest.main()



#
#  Copyright (C) 2003  Greg Landrum and Rational Discovery LLC
#
"""unit testing code for the SimilarityScreeners

"""
import unittest

from rdkit import Chem, DataStructs
from rdkit.Chem.Fingerprints import SimilarityScreener


def fingerprinter(mol):
  return Chem.RDKFingerprint(mol, minPath=2, maxPath=7, fpSize=2048)


class TestCase(unittest.TestCase):

  def test_SimilarityScreener(self):
    probe = fingerprinter(Chem.MolFromSmiles('C1OCCCC1'))

    screener = SimilarityScreener.SimilarityScreener()
    self.assertEqual(screener.probe, None)
    screener.SetProbe(probe)
    self.assertEqual(screener.probe, probe)

    screener = SimilarityScreener.SimilarityScreener(probe=probe, fingerprinter=fingerprinter)
    self.assertEqual(screener.probe, probe)

    mol = Chem.MolFromSmiles('CCCN')
    self.assertEqual(fingerprinter(mol), screener.GetSingleFingerprint(mol))

    screener.Reset()

  def test1_TopNScreener(self):
    smis = ['C1CCCCC1', 'C1OCCCC1', 'C1NCCCC1', 'c1ccccc1', 'C1C(C)CCCC1', 'C1C(C)C(C)CCC1']
    suppl = Chem.SmilesMolSupplierFromText('\n'.join(smis), delimiter=",", smilesColumn=0,
                                           nameColumn=-1, titleLine=0)
    metric = DataStructs.TanimotoSimilarity
    mol = Chem.MolFromSmiles('C1OCCCC1')
    probe = fingerprinter(mol)

    screener = SimilarityScreener.TopNScreener(3, probe=probe, metric=metric,
                                               fingerprinter=fingerprinter, dataSource=suppl)
    self.assertIsNone(screener.topN)
    matches1 = [x for x in screener]
    self.assertIsNotNone(screener.topN)

    self.assertEqual(len(matches1), 3)
    matches2 = [x for x in screener]
    self.assertEqual(len(matches2), 3)
    self.assertEqual(matches1, matches2)

    self.assertEqual(probe, screener.GetSingleFingerprint(mol))

    # Getting the length also triggers the execution of the screen
    screener = SimilarityScreener.TopNScreener(3, probe=probe, metric=metric,
                                               fingerprinter=fingerprinter, dataSource=suppl)
    self.assertIsNone(screener.topN)
    self.assertEqual(len(screener), 3)
    self.assertIsNotNone(screener.topN)

    # as does accessing elements by index
    screener = SimilarityScreener.TopNScreener(3, probe=probe, metric=metric,
                                               fingerprinter=fingerprinter, dataSource=suppl)
    self.assertIsNone(screener.topN)
    screener[1]
    self.assertIsNotNone(screener.topN)

  def test2_ThresholdScreener(self):
    smis = ['C1CCCCC1', 'C1OCCCC1', 'C1NCCCC1', 'c1ccccc1', 'C1C(C)CCCC1', 'C1C(C)C(C)CCC1']
    suppl = Chem.SmilesMolSupplierFromText('\n'.join(smis), delimiter=",", smilesColumn=0,
                                           nameColumn=-1, titleLine=0)

    metric = DataStructs.TanimotoSimilarity
    probe = fingerprinter(Chem.MolFromSmiles('C1OCCCC1'))

    screener = SimilarityScreener.ThresholdScreener(0.09, probe=probe, metric=metric,
                                                    fingerprinter=fingerprinter, dataSource=suppl)
    matches1 = [x[0] for x in screener]
    self.assertEqual(len(matches1), 5)
    matches2 = [x[0] for x in screener]
    self.assertEqual(len(matches2), 5)
    self.assertEqual(matches1, matches2)

  def test3_ThresholdScreener_folding(self):
    smis = ['C1CCCCC1', 'C1OCCCC1', 'C1NCCCC1', 'c1ccccc1', 'C1C(C)CCCC1', 'C1C(C)C(C)CCC1']
    suppl = Chem.SmilesMolSupplierFromText('\n'.join(smis), delimiter=",", smilesColumn=0,
                                           nameColumn=-1, titleLine=0)

    metric = DataStructs.TanimotoSimilarity
    probe = Chem.RDKFingerprint(Chem.MolFromSmiles('C1OCCCC1'), minPath=2, maxPath=7, fpSize=4096)

    screener = SimilarityScreener.ThresholdScreener(0.09, probe=probe, metric=metric,
                                                    fingerprinter=fingerprinter, dataSource=suppl)
    matches1 = [x[0] for x in screener]
    self.assertEqual(len(matches1), 5)
    matches2 = [x[0] for x in screener]
    self.assertEqual(len(matches2), 5)
    self.assertEqual(matches1, matches2)


if __name__ == '__main__':  # pragma: nocover
  unittest.main()
